"""Definition of the plane class.

This module defines the plane object in the Batoms package.

"""

import bpy
from time import time
import numpy as np
from batoms.base.object import BaseObject
from .setting import MolecularSurfaceSettings
import logging

logger = logging.getLogger(__name__)


class MolecularSurface(BaseObject):
    def __init__(
        self,
        label=None,
        location=np.array([0, 0, 0]),
        batoms=None,
    ):
        """Plane Class

        Args:
            label (_type_, optional): _description_. Defaults to None.
            location (_type_, optional): _description_. Defaults to np.array([0, 0, 0]).
            batoms (_type_, optional): _description_. Defaults to None.
        """
        #
        self.batoms = batoms
        self.label = label
        name = "plane"
        BaseObject.__init__(self, label, name)
        self.settings = MolecularSurfaceSettings(self.label, parent=self)
        self.settings.bpy_data.active = True

    def build_materials(
        self,
        name,
        color,
        node_inputs=None,
        material_style="default",
        vertex_color=None,
        color_by_attribute=None,
    ):
        """ """
        from batoms.material import create_material

        if name in bpy.data.materials:
            mat = bpy.data.materials.get(name)
            bpy.data.materials.remove(mat, do_unlink=True)
        mat = create_material(
            name,
            color=color,
            node_inputs=node_inputs,
            material_style=material_style,
            backface_culling=False,
            vertex_color=vertex_color,
            color_by_attribute=color_by_attribute,
        )
        return mat

    def draw(self, ms_name="ALL"):
        from batoms.utils.butils import clean_coll_object_by_type

        # delete old plane
        clean_coll_object_by_type(self.batoms.coll, "MS")
        for ms in self.settings.bpy_setting:
            if ms_name.upper() != "ALL" and ms.name != ms_name:
                continue
            if ms.type == "SAS":
                self.draw_SAS(ms)
            elif ms.type == "SES":
                self.draw_SES(ms)

    @property
    def sas_objs(self):
        sas_objs = {}
        for ms in self.settings.bpy_setting:
            sas_name = "%s_%s_sas" % (self.label, ms.name)
            obj = bpy.data.objects.get(sas_name)
            sas_objs[ms.name] = obj
        return sas_objs

    @property
    def sas_mb(self):
        return bpy.data.metaballs.get(self.sas_name)

    @property
    def sas_mat(self):
        return bpy.data.materials.get(self.sas_name)

    @property
    def ses_objs(self):
        ses_objs = {}
        for ms in self.settings.bpy_setting:
            sas_name = "%s_%s_ses" % (self.label, ms.name)
            obj = bpy.data.objects.get(sas_name)
            ses_objs[ms.name] = obj
        return ses_objs

    def get_space(self, resolution):
        return [resolution] * 3

    def get_box(self, vertices, padding=4):
        """
        Find a minmum box contains all atoms
        """
        from batoms.utils import get_box

        self.box = get_box(vertices, padding=padding)
        self.box_origin = self.box[:, 0]
        return self.box, self.box_origin

    def build_grid(self, resolution):
        """
        generate gridpoints
        """
        from batoms.utils import build_grid

        self.meshgrids, self.shape = build_grid(self.box, resolution)

    def calc_isosurface(
        self, volume, level, spacing, origin=[0, 0, 0], color=(0.2, 0.1, 0.9, 1.0)
    ):
        """
        Computes an isosurface from a volume grid.
        Parameters:
        volume: np.array
        cell: np.array
        level: float
        """
        from skimage import measure

        tstart = time()
        verts, faces, normals, values = measure.marching_cubes(
            volume, level=level, spacing=spacing
        )
        verts += origin
        faces = list(faces)
        isosurface = {
            "vertices": verts,
            "edges": [],
            "faces": faces,
            "color": color,
            "battr_inputs": {"Bmolecularsurface": {}},
        }
        logger.debug("Marching_cube: %s" % (time() - tstart))
        return isosurface

    def draw_SAS(self, ms, parallel=1):
        """
        1) Generate meshgrids
        2) Calculate power distance for grids
        3) Marching cube find isosurface = 5
        """
        from batoms.draw import draw_surface_from_vertices

        resolution = ms.resolution
        probe = ms.probe
        logger.debug("Resolution: {:1.3f}, Probe: {:1.3}".format(resolution, probe))
        tstart = time()
        indices = self.batoms.selects[ms.select].indices
        if len(indices) == 0:
            return
        radii = np.array(self.batoms.radii_vdw[indices]) + probe
        positions = self.batoms.positions[indices]
        self.get_box(positions, padding=max(radii) + resolution)
        self.build_grid(resolution=resolution)
        logger.debug("Grid Points: %s %s %s" % self.shape)
        indices_sas, volume_sas = self.calc_power_distance(
            self.meshgrids, positions, radii, parallel=parallel
        )
        volume = volume_sas.reshape(self.shape)
        isosurface = self.calc_isosurface(
            volume, 5, spacing=self.get_space(resolution), origin=self.box_origin
        )
        isosurface["color"] = ms.color
        isosurface["material_style"] = ms.material_style
        logger.debug("Vertices: %s" % len(isosurface["vertices"]))
        sas_name = "%s_%s_sas" % (self.label, ms.name)
        self.delete_obj(sas_name)
        coll = self.batoms.coll.children["%s_surface" % self.batoms.coll_name]
        obj = draw_surface_from_vertices(
            sas_name,
            datas=isosurface,
            coll=coll,
        )
        #
        if ms.color_by != "None":
            from ase.cell import Cell
            from batoms.utils import map_volumetric_data
            from batoms.utils.attribute import set_mesh_attribute

            if ms.color_by.upper() == "ELECTROSTATIC_POTENTIAL":
                if "charges" not in self.batoms.obj.data.attributes:
                    self.batoms.auto_assign_charge()
                data = self.batoms.calc_electrostatic_potential(isosurface["vertices"])
            else:
                # using volumetric_data
                # scaled positions
                scaled_verts = Cell(self.batoms.cell).scaled_positions(
                    isosurface["vertices"]
                )
                data = map_volumetric_data(
                    self.batoms.volumetric_data[ms.color_by], scaled_verts
                )
            # normalize
            data = (data - np.min(data)) / (np.max(data) - np.min(data))
            # color_attribute = map_color(data, [1, 0, 0, ms.transparency],
            # [0, 0, 1, ms.transparency])
            # set_vertex_color(obj, ms.color_by, color_attribute)
            obj.data.attributes.new(
                name="{}_data".format(ms.color_by), type="FLOAT", domain="POINT"
            )
            set_mesh_attribute(obj, "{}_data".format(ms.color_by), data)
            color_by_attribute = {
                "attribute_name": "{}_data".format(ms.color_by),
                "ValToRGB": [ms.color1[:], ms.color2[:]],
            }
        else:
            color_by_attribute = None
        mat = self.build_materials(
            sas_name,
            color=isosurface["color"],
            material_style=isosurface["material_style"],
            #    vertex_color = ms.color_by,
            color_by_attribute=color_by_attribute,
        )
        obj.data.materials.append(mat)

        obj.parent = self.batoms.obj
        obj.batoms.type = "MS"
        obj.batoms.label = self.batoms.label
        logger.debug("Draw SAS: %s" % (time() - tstart))

    def draw_SES(self, ms, parallel=1):
        """
        1) Get SAS
        2) Point probe sphere on vertices of SAS mesh
        3) Calculate power distance of grids
        4) value outside SAS set a value smaller than 5
        5) Marching cube find isosurface = 5
        """
        from batoms.draw import draw_surface_from_vertices

        resolution = ms.resolution
        probe = ms.probe
        logger.debug("Resolution: {:1.3f}, Probe: {:1.3}".format(resolution, probe))
        tstart = time()
        indices = self.batoms.selects[ms.select].indices
        if len(indices) == 0:
            return
        radii_vdw = np.array(self.batoms.radii_vdw[indices])
        radii = radii_vdw + probe
        positions = self.batoms.positions[indices]
        self.get_box(positions, padding=max(radii) + resolution + probe)
        self.build_grid(resolution=resolution)
        # draw_vertices('meshgrid', self.meshgrids)
        logger.debug("Grid Points: %s %s %s" % self.shape)
        # build SAS
        indices_sas, volume_sas = self.calc_power_distance(
            self.meshgrids, positions, radii, parallel=parallel
        )
        volume = volume_sas.reshape(self.shape)
        isosurface = self.calc_isosurface(
            volume, 5, self.get_space(resolution), origin=self.box_origin
        )
        # build SAS with probe = -self.resolution
        radii = radii_vdw - resolution
        volume_sas1 = np.where(volume_sas > 5, 6, volume_sas)
        indices = np.where(volume_sas < 5)[0]
        volume_sas1[indices]
        # select grid points inside sas for query
        indices_sas1, tmp = self.calc_power_distance(
            self.meshgrids[indices], positions, radii, parallel=parallel
        )
        volume_sas1[indices] = tmp
        # ----------------------------
        # tstart1 = time()
        vertices = isosurface["vertices"]
        nvert = len(vertices)
        radii = np.ones(nvert) * probe
        # select grid points between sas and sas1
        self.volume_ses = np.where(volume_sas > 5, 4, volume_sas)
        self.volume_ses = np.where(volume_sas1 < 5, 6, self.volume_ses)
        indices = np.where((volume_sas < 5) & (volume_sas1 > 5))[0]
        indices_ses, volume_ses = self.calc_power_distance(
            self.meshgrids[indices], vertices, radii, parallel=parallel
        )
        self.volume_ses[indices] = volume_ses
        volume = self.volume_ses.reshape(self.shape)
        isosurface = self.calc_isosurface(
            volume, 5, self.get_space(resolution), origin=self.box_origin
        )
        isosurface["color"] = ms.color
        isosurface["material_style"] = ms.material_style
        logger.debug("Vertices: %s" % len(isosurface["vertices"]))
        ses_name = "%s_%s_ses" % (self.label, ms.name)
        self.delete_obj(ses_name)
        coll = self.batoms.coll.children["%s_surface" % self.batoms.coll_name]
        obj = draw_surface_from_vertices(
            ses_name,
            datas=isosurface,
            coll=coll,
        )
        mat = self.build_materials(
            ses_name,
            color=isosurface["color"],
            material_style=isosurface["material_style"],
        )
        obj.data.materials.append(mat)
        obj.parent = self.batoms.obj
        obj.batoms.type = "MS"
        obj.batoms.label = self.batoms.label
        logger.debug("Time SES: %s" % (time() - tstart))
        self.get_sesa(ms.name)

    def get_sasa(self, name):
        """ """
        from batoms.utils.butils import get_area, get_volume

        me = self.sas_objs[name].data
        area = get_area(me)
        volume = get_volume(me)
        print("Area: {:5.3f},    Volume: {:5.3f}".format(area, volume))
        return area, volume

    def get_psasa(self):
        """ """
        me = self.sas_objs.data
        npoly = len(me.polygons)
        areas = np.zeros(npoly)
        me.polygons.foreach_get("area", areas)
        pareas = {}
        arrays = self.batoms.arrays
        positions = arrays["positions"]
        symbols = arrays["species"]
        natom = len(positions)
        radii = np.array(self.batoms.radii_vdw) + self.probe
        atom_indices, distance = self.map_face_to_atom(me, positions, radii)
        for j in range(natom):
            parea = np.sum(areas[atom_indices == j])
            pareas[j] = [symbols[j], parea]
            print("species: {}, area: {:1.3f}".format(symbols[j], parea))
        # print(pareas_list)
        return pareas

    def get_sesa(self, name):
        """ """
        from batoms.utils.butils import get_area, get_volume

        me = self.ses_objs[name].data
        area = get_area(me)
        volume = get_volume(me)
        print("SES: area: {:1.3f}, volume: {:1.3f}".format(area, volume))
        return area, volume

    def to_mesh(self):
        """ """
        # tstart = time()
        depsgraph = bpy.context.evaluated_depsgraph_get()
        me = self.sas_objs.evaluated_get(depsgraph).to_mesh()
        nvert = len(me.vertices)
        npoly = len(me.polygons)
        logger.debug("vertices: %s, polygons: %s." % (nvert, npoly))
        # print('SAS to mesh evaluated: %s' % (time() - tstart))
        return me

    def get_sasa_mb(self, frame_indices=[]):
        """ """
        from batoms.utils.butils import get_area, get_volume

        if isinstance(frame_indices, int):
            frame_indices = [frame_indices]
        if frame_indices == []:
            frame_indices = range(self.nframe)
        areas_list = []
        pareas_list = []
        tstart = time()
        print("Frame    SAS Area    SAS Volume")
        for i in frame_indices:
            bpy.context.scene.frame_set(i)
            me = self.to_mesh()
            area = get_area(me)
            volume = get_volume(me)
            print("{:4d}  {:11.3f}  {:11.3f}".format(i, area, volume))
            areas_list.append(area)
        logger.debug("Time for SAS area: %s" % (time() - tstart))
        return areas_list, pareas_list

    def get_psasa_mb(self, frame_indices=[]):
        """ """
        if isinstance(frame_indices, int):
            frame_indices = [frame_indices]
        if frame_indices == []:
            frame_indices = range(self.nframe)
        pareas_list = []
        radii = np.array(self.batoms.radii_vdw) + self.probe
        for i in frame_indices:
            bpy.context.scene.frame_set(i)
            me = self.to_mesh()
            npoly = len(me.polygons)
            areas = np.zeros(npoly)
            me.polygons.foreach_get("area", areas)
            pareas = {}
            atoms = self.batoms.atoms
            positions = atoms.positions
            symbols = atoms.get_chemical_symbols()
            natom = len(positions)
            atom_indices, distance = self.map_face_to_atom(me, positions, radii)
            for j in range(natom):
                parea = np.sum(areas[atom_indices == j])
                pareas[j] = [symbols[j], parea]
            pareas_list.append(pareas)
        # print(pareas_list)
        return pareas_list

    def map_face_to_atom(self, me, positions, radii, k=1):
        """ """
        npoly = len(me.polygons)
        centers = np.zeros(npoly * 3)
        me.polygons.foreach_get("center", centers)
        centers = centers.reshape(-1, 3)
        indices, distance = self.calc_power_distance(centers, positions, radii, k=k)
        return indices, distance

    def map_vertice_to_atom(self, me, positions, radii, k=1):
        """ """
        nvertice = len(me.vertice)
        vertices = np.zeros(nvertice * 3)
        me.vertices.foreach_get("co", vertices)
        vertices = vertices.reshape(-1, 3)
        indices, distance = self.calc_power_distance(vertices, positions, radii, k=k)
        return indices, distance

    def query_radius(self, points, positions, radii, k=1):
        """
        Algorithm:
        Use KDTree to query the tree for neighbors within a radius r.
        """
        from scipy import spatial

        tstart = time()
        # ----------------------------------------------------
        tstart = time()
        tree = spatial.KDTree(points)
        logger.debug("KDTree point: %s" % (time() - tstart))
        tstart = time()
        indices = tree.query_ball_point(positions, radii)
        logger.debug("KDTree query: %s" % (time() - tstart))
        tstart = time()
        indices = np.unique(np.concatenate(indices).astype(int))
        logger.debug("array indices: %s" % (time() - tstart))
        return indices

    def calc_power_distance(self, points, positions, radii, k=1, parallel=1):
        """
        Algorithm:
        Use KDTree to find the nearest atom for all vertices with
        the power distance as metric.
        """
        from scipy import spatial

        tstart = time()
        npoint = len(points)
        # ----------------------------------------------------
        # change to 4-dimension to set power distance
        points = np.append(points, np.zeros((npoint, 1)), axis=1)
        # maxr = max(max(radii), 5)
        maxr = 5
        radii = np.sqrt(maxr * maxr - np.square(radii)).reshape(-1, 1)
        positions = np.append(positions, radii, axis=1)
        tstart = time()
        tree = spatial.KDTree(positions)
        logger.debug("KDTree positions: %s" % (time() - tstart))
        tstart = time()
        distance, indices = tree.query(points, k=k, workers=parallel)
        logger.debug("KDTree query: %s" % (time() - tstart))
        return indices, distance

    def to_mesh_object(self):
        """ """
        import bmesh

        mbmesh = self.to_mesh()
        bpy.data.objects.remove(self.sas_objs, do_unlink=True)
        bpy.data.metaballs.remove(self.sas_mb, do_unlink=True)
        bm = bmesh.new()
        bm.from_mesh(mbmesh)
        tmp = bpy.data.meshes.new(name="Mesh")
        bm.to_mesh(tmp)
        obj = bpy.data.objects.new(self.sas_name, tmp)
        obj.data.materials.append(self.sas_mat)
        # print('SAS to mesh evaluated: %s'%(time() - tstart))
        coll = self.batoms.coll.children["%s_surface" % self.batoms.coll_name]
        coll.objects.link(obj)
        return obj

    def draw_SES_mb(
        self,
        resolution=0.4,
        threshold=1e-4,
        stiffness=1,
        area=True,
        smooth=None,
        subdivide=0,
        refine_normal=True,
        refine=True,
        steps=5,
    ):
        """
        Time:
        sas_mesh: 1.1
        shift_vertices: 0.09
        vertices_smooth: 0.8
        total: 2.55
        """
        name = "%s_ses" % self.label
        tstart0 = time()
        positions = self.batoms.positions
        radii = np.array(self.batoms.radii_vdw) + self.probe
        obj = self.build_SAS(
            resolution=resolution, threshold=threshold, stiffness=stiffness
        )
        obj = self.to_mesh_object()
        logger.debug("sas_mesh: %s" % (time() - tstart0))
        obj.name = name
        eps = resolution / 2
        # -------------------------------------
        self.SES_subdivide(positions, radii, eps, subdivide)
        # return
        # ----------------------
        me = self.ses_obj.data
        nvert = len(me.vertices)
        vertices = np.zeros(nvert * 3)
        normals = np.zeros(nvert * 3)
        me.vertices.foreach_get("co", vertices)
        me.vertices.foreach_get("normal", normals)
        vertices = vertices.reshape(-1, 3)
        normals = normals.reshape(-1, 3)
        indices, mask1, mask2, mask3 = self.SES_map_to_atom(
            vertices, positions, radii, eps
        )
        # return
        # refine_normal = False
        if refine_normal:
            normals = self.SES_refine_normal(
                vertices, normals, positions, indices, mask1, mask2, mask3
            )
        vertices = vertices - normals * self.probe
        vertices = vertices.reshape(-1, 1)
        me.vertices.foreach_set("co", vertices)
        # return
        # ----------------------------------
        mask = np.ones(nvert, dtype=bool)
        if refine:
            mask = ~mask1
            for i in range(steps):
                self.SES_vertices_smooth(
                    mask, resolution=resolution, subdivide=0, smooth=3
                )
                self.SES_refine_position(positions, indices, mask1, mask2, mask3)
        self.SES_vertices_smooth(
            mask, resolution=resolution, subdivide=subdivide, smooth=smooth
        )
        logger.debug("Time: %s" % (time() - tstart0))
        if area:
            self.get_sesa()
        return obj

    def SES_map_to_atom(self, vertices, positions, radii, eps, obj=None):
        """
        map vertices to atoms
        1) three sphere
        2) two sphere
        1) one sphere: contact surface
        """
        from batoms.utils import check_origin_3, check_origin_2
        from batoms.draw import draw_vertices

        logger.debug("eps: ", eps)
        n = len(vertices)
        indices, distance = self.calc_power_distance(vertices, positions, radii, k=3)
        delta0 = abs(distance[:, 0] - distance[:, 1])
        delta1 = abs(distance[:, 1] - distance[:, 2])
        delta2 = abs(distance[:, 0] - distance[:, 2])
        # three spheres
        indices3 = np.where((delta0 < eps) & (delta1 < eps) & (delta2 < eps))[0]
        origins0 = positions[indices[:, 0][indices3]]
        origins1 = positions[indices[:, 1][indices3]]
        origins2 = positions[indices[:, 2][indices3]]
        vertices3 = vertices[indices3]
        indices31, origins_probe3 = check_origin_3(
            vertices3,
            origins0,
            origins1,
            origins2,
            radii[indices[:, 0][indices3]],
            radii[indices[:, 1][indices3]],
            radii[indices[:, 2][indices3]],
            0,
            5 * eps,
        )
        indices3 = indices3[indices31]
        mask3 = np.zeros(n, dtype=bool)
        mask3[indices3] = True
        draw_vertices("origin3", origins_probe3)
        # two spheres
        indices2 = np.where((delta0 < eps))[0]
        origins0 = positions[indices[:, 0][indices2]]
        origins1 = positions[indices[:, 1][indices2]]
        vertices2 = vertices[indices2]
        indices21, origins_probe2 = check_origin_2(
            vertices2,
            origins0,
            origins1,
            radii[indices[:, 0][indices2]],
            radii[indices[:, 1][indices2]],
            0,
            5 * eps,
        )
        indices2 = indices2[indices21]
        mask2 = np.zeros(n, dtype=bool)
        mask2[indices2] = True
        draw_vertices("origin2", origins_probe2)
        # ---------------------------
        mask1 = np.where(delta0 > eps, True, False)
        logger.debug(
            "indices: 1, 2, 3 ",
            len(indices),
            np.count_nonzero(mask1),
            np.count_nonzero(mask2),
            np.count_nonzero(mask3),
        )
        # ----------------
        self.origins_probe2 = origins_probe2
        self.origins_probe3 = origins_probe3
        if obj is not None:
            self.batoms_mesh(obj, mask3, mode="VERT")
        return indices, mask1, mask2, mask3

    def select_mesh(self, obj, mask, mode="VERT"):
        bpy.context.view_layer.objects.active = obj
        bpy.ops.object.mode_set(mode="EDIT")
        bpy.ops.mesh.select_all(action="DESELECT")
        bpy.ops.object.mode_set(mode="OBJECT")
        if mode == "VERT":
            obj.data.vertices.foreach_set("select", mask)
        elif mode == "EDGE":
            obj.data.edges.foreach_set("select", mask)
        elif mode == "FACE":
            obj.data.polygons.foreach_set("select", mask)
        bpy.ops.object.mode_set(mode="EDIT")
        bpy.ops.mesh.select_mode(type=mode)

    def SES_subdivide(self, positions, radii, eps, number_cuts=2):
        """
        subdivide edge
        """
        if number_cuts < 1:
            return
        me = self.ses_obj.data
        nedge = len(me.edges)
        nvert = len(me.vertices)
        vertices = np.zeros(nvert * 3)
        iverts = np.zeros(nedge * 2, dtype="int")
        me.vertices.foreach_get("co", vertices)
        me.edges.foreach_get("vertices", iverts)
        vertices = vertices.reshape(-1, 3)
        iverts = iverts.reshape(-1, 2)
        centers = (vertices[iverts[:, 0]] + vertices[iverts[:, 1]]) / 2
        indices, distance = self.calc_power_distance(centers, positions, radii, k=3)
        delta0 = abs(distance[:, 0] - distance[:, 1])
        delta1 = abs(distance[:, 0] - distance[:, 2])
        delta0 = np.absolute(distance[:, 0] - distance[:, 1])
        normals1 = vertices[iverts[:, 0]] - vertices[iverts[:, 1]]
        lengths = np.linalg.norm(normals1, axis=1)
        normals1 = normals1 / lengths[:, None]
        normals2 = positions[indices[:, 0]] - positions[indices[:, 1]]
        normals2 = normals2 / np.linalg.norm(normals2, axis=1)[:, None]
        # proj = np.absolute(np.einsum('ij,ij->i', normals1, normals2))
        mask = np.where(
            (((delta0 < eps)) | (delta1 < eps)) & (lengths > eps / 2 / number_cuts),
            True,
            False,
        )
        # print(eps, lengths[2608], mask[2608])
        self.batoms_mesh(self.ses_obj, mask, mode="EDGE")
        bpy.ops.mesh.subdivide(
            number_cuts=int(number_cuts), smoothness=0, fractal_along_normal=0
        )
        # return
        bpy.ops.object.mode_set(mode="OBJECT")

    def SES_refine_normal(
        self, vertices, normals, positions, indices, indices1, indices2, indices3
    ):
        # -----------------------------------------
        origins0 = positions[indices[:, 0][indices1]]
        vertices0 = vertices[indices1]
        vecs0 = vertices0 - origins0
        vecs0 = vecs0 / np.linalg.norm(vecs0, axis=1)[:, None]
        normals[indices1] = vecs0
        # -------------------
        origins0 = positions[indices[:, 0][indices2]]
        origins1 = positions[indices[:, 1][indices2]]
        vertices1 = vertices[indices2]
        xaxis = origins1 - origins0
        l2 = np.linalg.norm(xaxis, axis=1)
        # Heron's formula
        xaxis = xaxis / l2[:, None]
        vecs1 = vertices1 - origins0
        zaxis = np.cross(xaxis, vecs1)
        zaxis = zaxis / np.linalg.norm(zaxis, axis=1)[:, None]
        projz = np.einsum("ij,ij->i", normals[indices2], zaxis)[:, None] * zaxis
        normals[indices2] = normals[indices2] - projz
        return normals

    def SES_vertices_smooth(self, mask, resolution=0.4, subdivide=0, smooth=None):
        if smooth is None:
            repeat = int(1 / resolution)
        else:
            repeat = smooth
        if repeat == 0:
            return
        repeat += subdivide
        tstart = time()
        bpy.context.view_layer.objects.active = self.ses_obj
        self.batoms_mesh(self.ses_obj, mask, mode="VERT")
        if repeat > 0:
            bpy.ops.mesh.vertices_smooth(factor=0.5, repeat=repeat)
        bpy.ops.object.mode_set(mode="OBJECT")
        logger.debug("smooth_vertices: %s" % (time() - tstart))

    def SES_refine_position(self, positions, indices, indices1, indices2, indices3):
        radii_vdw = np.array(self.batoms.radii_vdw)
        obj = self.ses_obj
        me = obj.data
        # -------------------------------------------------------
        tstart = time()
        nvert = len(me.vertices)
        vertices = np.zeros(nvert * 3)
        # print('vertices: %s'%(nvert))
        me.vertices.foreach_get("co", vertices)
        vertices = vertices.reshape(-1, 3)
        # -----------------------------------------
        # print('indices1: ', len(indices1))
        if True:
            origins0 = positions[indices[:, 0][indices1]]
            vertices1 = vertices[indices1]
            vecs0 = vertices1 - origins0
            dist = np.linalg.norm(vecs0, axis=1)[:, None]
            radii_vdw0 = radii_vdw[indices[:, 0][indices1]]
            normals0 = vecs0 / dist
            vertices[indices1] = origins0 + radii_vdw0[:, None] * normals0
        # ----------
        # print('indices2: ', len(indices2))
        vertices2 = vertices[indices2]
        vecs2 = vertices2 - self.origins_probe2
        dist = np.linalg.norm(vecs2, axis=1)[:, None]
        normals2 = vecs2 / dist
        vertices[indices2] = self.origins_probe2 + self.probe * normals2
        # draw_vertices('origin2', origins_probe)
        # ----------------------------------------------------------
        vertices3 = vertices[indices3]
        vecs3 = vertices3 - self.origins_probe3
        dist = np.linalg.norm(vecs3, axis=1)[:, None]
        normals3 = vecs3 / dist
        vertices[indices3] = self.origins_probe3 + self.probe * normals3
        # draw_vertices('origin3', origins_probe)
        # ----------------------------------------------------------
        vertices = vertices.reshape(-1, 1)
        me.vertices.foreach_set("co", vertices)
        me.update()
        logger.debug("refine vertices: %s" % (time() - tstart))

    def get_sesa_mb(self):
        """ """
        from batoms.utils.butils import get_area, get_volume

        me = self.ses_obj.data
        area = get_area(me)
        volume = get_volume(me)
        print("SES: area: {:1.3f}, volume: {:1.3f}".format(area, volume))
        return area, volume

    def build_SAS_Shrake_Rupley(self, probe=1.4, subdivisions=2):
        """
        Algorithm: ShrakeRupley
        """
        default_colors = [(0.2, 0.1, 0.9, 1.0), (0.0, 0.0, 1.0, 1.0)]
        #
        tstart = time()
        self.probe = probe
        atoms = self.batoms
        positions = atoms.positions
        radii = np.array(self.batoms.radii_vdw) + self.probe
        na = len(positions)
        #
        tstart = time()
        bpy.ops.mesh.primitive_ico_sphere_add(subdivisions=subdivisions)
        # bpy.ops.mesh.primitive_cylinder_add()
        source = bpy.context.view_layer.objects.active
        me = source.data
        n = len(me.vertices)
        sphere0 = np.empty(n * 3, dtype=np.float64)
        me.vertices.foreach_get("co", sphere0)
        sphere0 = sphere0.reshape((n, 3))
        spheres = np.tile(sphere0, (na, 1, 1))
        #
        radii = np.array(radii).reshape(-1, 1)
        scales = radii
        scales = np.concatenate((scales, scales, scales), axis=1)
        spheres = spheres * scales[:, None]
        #
        origini = positions
        spheres += origini[:, None]
        logger.debug("build_spheres: {0:10.2f} s".format(time() - tstart))
        #
        spheres = spheres.reshape(-1, 3)
        indices, distances = self.calc_power_distance(spheres, positions, radii)
        mask = np.where(distances > 4.98)
        spheres = spheres[mask]
        #
        # tstart1 = time()
        faces = []
        """
        # tri = spatial.Delaunay(spheres)
        # faces = tri.simplices
        tree = spatial.KDTree(spheres)
        logger.debug('KDTree positions: %s'%(time() - tstart))
        distance, indices = tree.query(spheres, k = 3)
        indices0 = np.arange(len(spheres))
        indices0 = indices0.reshape(-1, 1)
        faces = np.concatenate((indices0, indices[:, 1:]), axis = 1)
        faces = faces.tolist()
        logger.debug('build_faces: %s'%(time() - tstart1))
        """
        #
        SAS_Shrake_Rupley = {
            "vertices": spheres.reshape(-1, 3),
            "edges": [],
            # 'faces': [],
            "faces": faces,
            # 'faces': list(tri.simplices),
            "color": default_colors[0],
            "battr_inputs": {},
        }
        logger.debug("vertices: ", len(SAS_Shrake_Rupley["vertices"]))
        logger.debug("build_SAS: %s" % (time() - tstart))
        self.SAS_Shrake_Rupley = SAS_Shrake_Rupley
        bpy.data.objects.remove(source, do_unlink=True)
        # return 0

    @property
    def setting(self):
        from batoms.utils import deprecated

        """setting object."""
        deprecated(
            '"setting" will be deprecated in the furture, please use "settings".'
        )
        return self.settings

    def as_dict(self):
        """ """
        data = {}
        data["settings"] = self.settings.as_dict()
        data.update(self.settings.bpy_data.as_dict())
        return data



from bpy.props import BoolProperty, StringProperty
from batoms import Batoms
from batoms.ops.base import OperatorBatoms


class TemplateAdd(OperatorBatoms):
    bl_idname = "template.template_add"
    bl_label = "Add template"
    bl_description = "Add template to a Batoms"

    name: StringProperty(
        name="name", default="2", description="Name of template to be added"
    )

    def execute(self, context):
        obj = context.object
        batoms = Batoms(label=context.object.batoms.label)
        batoms.template.settings.add(self.name)
        context.view_layer.objects.active = obj
        return {"FINISHED"}


class TemplateRemove(OperatorBatoms):
    bl_idname = "template.template_remove"
    bl_label = "Remove template"
    bl_description = "Remove template to a Batoms"

    name: StringProperty(
        name="name", default="1-1-1", description="Name of template to be removed"
    )

    all: BoolProperty(name="all", default=False, description="Remove all template")

    def execute(self, context):
        obj = context.object
        batoms = Batoms(label=obj.batoms.label)
        batoms.template.settings.remove((self.name))
        context.view_layer.objects.active = obj
        return {"FINISHED"}


class TemplateDraw(OperatorBatoms):
    bl_idname = "template.template_draw"
    bl_label = "Draw template"
    bl_description = "Draw template to a Batoms"

    name: StringProperty(
        name="name", default="ALL", description="Name of template surface to be drawed"
    )

    def execute(self, context):
        obj = context.object
        batoms = Batoms(label=obj.batoms.label)
        batoms.template.draw()
        context.view_layer.objects.active = batoms.obj
        return {"FINISHED"}



import bpy
from batoms import Batoms


def test_batoms_delete():
    """ """
    from batoms import Batoms

    bpy.ops.batoms.delete()
    # delete all
    bpy.ops.batoms.molecule_add(label="nh3", formula="NH3")
    bpy.ops.batoms.molecule_add(label="h2o", formula="H2O")
    bpy.ops.batoms.delete()
    assert "nh3" not in bpy.data.collections
    assert "h2o" not in bpy.data.collections
    # delete by label
    bpy.ops.batoms.molecule_add(label="nh3", formula="NH3")
    bpy.ops.batoms.molecule_add(label="h2o", formula="H2O")
    bpy.ops.batoms.delete(label="h2o")
    assert "h2o" not in bpy.data.collections
    # delete by selection
    bpy.ops.batoms.molecule_add(label="nh3", formula="NH3")
    bpy.ops.batoms.molecule_add(label="h2o", formula="H2O")
    nh3 = Batoms("nh3")
    h2o = Batoms("h2o")
    nh3.obj.select_set(False)
    h2o.obj.select_set(True)
    bpy.ops.batoms.delete_selected_batoms()
    assert "h2o" not in bpy.data.collections
    bpy.ops.batoms.delete()


def test_batoms_apply_model_style(h2o):
    bpy.context.view_layer.objects.active = h2o.obj
    assert h2o.model_style == 0
    # model_style
    bpy.ops.batoms.apply_model_style(model_style="1")
    assert h2o.model_style == 1
    # radius_style
    bpy.ops.batoms.apply_radius_style(radius_style="1")
    assert h2o.radius_style == "1"
    # color_style
    bpy.ops.batoms.apply_color_style(color_style="2")
    assert h2o.color_style == "2"


def test_batoms_join_seperate(h2o, ch4):
    """ """
    ch4.obj.select_set(True)
    h2o.obj.select_set(True)
    # join
    bpy.ops.batoms.join(label="ch4")
    assert len(ch4) == 8
    # separate
    ch4.separate()
    assert len(ch4) == 5
    assert len(h2o) == 3
    # join with another name
    ch4.obj.select_set(True)
    h2o.obj.select_set(True)
    bpy.ops.batoms.join(label="a")
    a = Batoms("a")
    assert len(a) == 8
    # separate
    a.separate()
    assert len(ch4) == 5
    assert len(h2o) == 3


def test_batoms_apply_label(h2o):
    bpy.context.view_layer.objects.active = h2o.obj
    bpy.ops.batoms.apply_label(label="elements")
    assert h2o.show_label == "elements"
    bpy.ops.batoms.apply_label(label="")
    assert h2o.show_label == ""


def test_ase_molecule():
    """Create a molecule use GUI ASE"""
    bpy.ops.batoms.delete()
    bpy.ops.batoms.molecule_add(label="nh3", formula="NH3")
    nh3 = Batoms("nh3")
    assert len(nh3) == 4
    bpy.ops.batoms.delete()


def test_ase_bulk():
    """Create an bulk use GUI ASE"""
    bpy.ops.batoms.delete()
    bpy.ops.batoms.bulk_add(label="au", formula="Au")
    au = Batoms("au")
    assert len(au) == 1
    # surface
    au.obj.select_set(True)
    bpy.ops.batoms.surface_add()
    bpy.ops.batoms.delete()


def test_ase_surface():
    """Create an surface use GUI ASE"""
    bpy.ops.batoms.delete()
    # fcc
    bpy.ops.surface.fcc100_add(label="au100")
    au100 = Batoms("au100")
    assert len(au100) == 4
    bpy.ops.surface.fcc110_add(label="au110")
    au110 = Batoms("au110")
    assert len(au110) == 4
    bpy.ops.surface.fcc111_add(label="au111")
    au111 = Batoms("au111")
    assert len(au111) == 4
    bpy.ops.surface.fcc211_add(label="au211")
    au211 = Batoms("au211")
    assert len(au211) == 12
    bpy.ops.surface.fcc111_root_add(label="au111_root")
    au111_root = Batoms("au111_root")
    assert len(au111_root) == 12
    # bcc
    bpy.ops.surface.bcc100_add(label="fe100")
    fe100 = Batoms("fe100")
    assert len(fe100) == 4
    bpy.ops.surface.bcc110_add(label="fe110")
    fe110 = Batoms("fe110")
    assert len(fe110) == 4
    bpy.ops.surface.bcc111_add(label="fe111")
    fe111 = Batoms("fe111")
    assert len(fe111) == 4
    bpy.ops.surface.bcc111_root_add(label="fe111_root")
    fe111_root = Batoms("fe111_root")
    assert len(fe111_root) == 12
    # hcp
    bpy.ops.surface.hcp0001_add(label="ti0001")
    ti0001 = Batoms("ti0001")
    assert len(ti0001) == 4
    bpy.ops.surface.hcp10m10_add(label="ti10m10")
    ti10m10 = Batoms("ti10m10")
    assert len(ti10m10) == 16
    bpy.ops.surface.hcp0001_root_add(label="ti0001_root")
    ti0001_root = Batoms("ti0001_root")
    assert len(ti0001_root) == 12
    # diamond
    bpy.ops.surface.diamond100_add(label="c100")
    c100 = Batoms("c100")
    assert len(c100) == 4
    bpy.ops.surface.diamond111_add(label="c111")
    c111 = Batoms("c111")
    assert len(c111) == 4
    bpy.ops.batoms.delete()


# ==============================================
# Below for edit mode
# ==============================================
def test_batoms_apply_model_style_selected(h2o):
    import numpy as np
    from batoms import Batoms

    bpy.ops.surface.fcc111_add(label="au111", symbol="Au", size=(1, 1, 4))
    au111 = Batoms("au111")
    au111 += h2o
    au111.model_style = 0
    bpy.context.view_layer.objects.active = au111.obj
    # only select h2o
    au111.obj.data.vertices.foreach_set("select", [0, 0, 0, 0, 1, 1, 1])
    # change model_style for selected atoms
    bpy.ops.batoms.apply_model_style_selected(model_style="1")
    assert au111.get_attribute("model_style")[0] == 0
    assert au111.get_attribute("model_style")[-1] == 1
    assert np.isclose(au111.get_attribute("scale")[0], 1)
    assert np.isclose(au111.get_attribute("scale")[-1], 0.4)
    bpy.ops.batoms.delete()



"""
#TODO add panel for select
#TODO active radius_style
#TODO merge select and SliceBatoms
"""
import bpy
from batoms.base.collection import Setting
import numpy as np
from batoms.utils import string2Number

# from time import time
import logging

# logger = logging.getLogger('batoms')
logger = logging.getLogger(__name__)


class Select:
    def __init__(
        self,
        label,
        parent,
        name="all",
        batoms=None,
        indices=None,
    ) -> None:
        """_summary_

        Args:
            label (_type_): _description_
            name (str, optional): _description_. Defaults to 'all'.
            batoms (_type_, optional): _description_. Defaults to None.
            indices (_type_, optional): _description_. Defaults to None.

        Examples:

        >>>sel = batoms.selects[key]
        >>>sel.model_style = 1

        """
        self.label = label
        self.parent = parent
        self.name = name
        self.batoms = batoms
        self.coll_name = "%s_%s" % (label, name)
        if indices is not None:
            self.indices = indices

    def get_bpy_setting(self):
        if self.coll_name:
            coll = bpy.data.collections.get(self.coll_name)
            data = getattr(coll.batoms, self.name)
        else:
            raise KeyError("The collection property {} not exist!".format(self.name))
        return data.setting

    @property
    def positions(self):
        return self.batoms.positions[self.indices]

    @property
    def radii_vdw(self):
        return self.batoms.radii_vdw[self.indices]

    @property
    def model_style(self):
        return self.get_model_style()

    @model_style.setter
    def model_style(self, model_style):
        self.set_model_style(model_style)

    def get_model_style(self):
        model_style = self.batoms.arrays["model_style"]
        return model_style[self.indices]

    def set_model_style(self, model_style):
        if int(model_style) == 0:
            self.scale = 1
        elif int(model_style) == 1:
            self.scale = 0.4
        elif int(model_style) == 2:
            self.scale = 0.4
        elif int(model_style) == 3:
            self.scale = 0.0001
        model_style0 = self.batoms.attributes["model_style"]
        model_style0[self.indices] = model_style
        model_style = {"model_style": model_style0}
        self.batoms.set_attributes(model_style)
        self.batoms.draw()

    @property
    def radius_style(self):
        return self.get_radius_style()

    @radius_style.setter
    def radius_style(self, radius_style):
        self.set_radius_style(radius_style)

    def get_radius_style(self):
        return self.parent.bpy_setting[self.name].radius_style

    def set_radius_style(self, radius_style):
        self.parent.bpy_setting[self.name].radius_style = radius_style

    @property
    def show(self):
        return self.get_show()

    @show.setter
    def show(self, state):
        self.set_show(state)

    def get_show(self):
        return self.batoms.attributes["show"][self.indices]

    def set_show(self, show):
        show0 = self.batoms.show
        show0[self.indices] = show
        self.batoms.set_attributes({"show": show0})

    @property
    def mask(self):
        return self.get_mask()

    def get_mask(self):
        mask = self.batoms.arrays["select"]
        mask = np.where(mask == string2Number(self.name))[0]
        return mask

    @property
    def vg(self):
        vg = self.batoms.obj.vertex_groups.get(self.name)
        if vg is None:
            vg = self.batoms.obj.vertex_groups.new(name=self.name)
        return vg

    @property
    def indices(self):
        return self.get_indices()

    @indices.setter
    def indices(self, indices):
        self.set_indices(indices)

    def get_indices(self):
        vg_idx = self.vg.index
        obj = self.batoms.obj
        indices = [
            v.index
            for v in obj.data.vertices
            if vg_idx in [vg.group for vg in v.groups]
        ]
        return indices

    def set_indices(self, indices):
        if isinstance(indices, np.ndarray):
            indices = indices.tolist()
        self.vg.add(indices, 1.0, "ADD")
        select = self.batoms.attributes["select"]
        select[self.indices] = string2Number("all")
        select[indices] = string2Number(self.name)
        select = {"select": select}
        self.batoms.set_attributes(select)

    @property
    def scale(self):
        return self.get_scale()

    @scale.setter
    def scale(self, scale):
        self.set_scale(scale)

    def get_scale(self):
        scale = self.batoms.attributes["scale"][self.indices]
        return scale

    def set_scale(self, scale):
        """ """
        scale0 = self.batoms.attributes["scale"]
        scale0[self.indices] = scale
        # for species
        if isinstance(scale, dict):
            species = self.batoms.attributes["species"]
            for key, value in scale.items():
                scale0[np.where((species == key) & self.mask)] = value
            scale = scale0
        self.batoms.set_attributes({"scale": scale0})

    @property
    def color(self):
        return self.get_color()

    @color.setter
    def color(self, color):
        self.set_color(color)

    def get_color(self):
        """ """
        color = {}
        materials = self.materials
        main_elements = self.main_elements
        for sp in self.batoms.species:
            for node in materials[sp][main_elements[sp]].node_tree.nodes:
                if "Base Color" in node.inputs:
                    node_color = node.inputs["Base Color"].default_value[:]
                if "Alpha" in node.inputs:
                    Alpha = node.inputs["Alpha"].default_value
            color[sp] = [node_color[0], node_color[1], node_color[2], Alpha]
        return color

    def set_color(self, color):
        if len(color) == 3:
            color = [color[0], color[1], color[2], 1]
        self.materials[self.main_elements].diffuse_color = color
        for node in self.materials[self.main_elements].node_tree.nodes:
            if "Base Color" in node.inputs:
                node.inputs["Base Color"].default_value = color
            if "Alpha" in node.inputs:
                node.inputs["Alpha"].default_value = color[3]

    @property
    def radius(self):
        """
        radius depends on radius style, thus could be with select
        """
        radius = {}
        for sp in self.bpy_setting:
            radius[sp.name] = sp.radius
        return radius

    @property
    def species_props(self):
        return self.get_species_props()

    def get_species_props(self):
        species_props = {}
        radius = self.radius
        color = self.color
        for sp in self.species:
            species_props[sp] = {"radius": radius[sp], "color": color[sp]}
        return species_props

    def __repr__(self) -> str:
        s = "-" * 60 + "\n"
        s = "Name   model_style   radius_style     show \n"
        s += "{0:10s} {1:10s}   {2:10s}   {3:10s} \n".format(
            self.name, str(self.model_style[0]), self.radius_style[0], str(self.show[0])
        )
        s += "-" * 60 + "\n"
        return s


class Selects(Setting):
    """
    sel = batoms.select.add(name, index)
    sel = batoms.select.from_dict({name: index})
    sel.model_style = 1
    """

    def __init__(self, label, batoms=None) -> None:
        Setting.__init__(self, label, coll_name=label)
        self.label = label
        self.name = "batoms"
        self.batoms = batoms

    def get_ui_list_index(self):
        return self.bpy_data.ui_list_index_select

    def set_ui_list_index(self, value):
        self.bpy_data.ui_list_index_select = value

    def get_bpy_setting(self):
        if self.coll_name:
            coll = bpy.data.collections.get(self.coll_name)
            data = getattr(coll, self.name)
        else:
            raise KeyError("The collection property {} not exist!".format(self.name))
        return data.settings_select

    @property
    def selects(self):
        return self.get_selects()

    def get_selects(self):
        selects = {}
        for b in self.bpy_setting:
            selects[b.name] = Select(
                self.label, parent=self, name=b.name, batoms=self.batoms
            )
        for vg in self.batoms.obj.vertex_groups:
            if vg.name not in selects:
                self.add(vg.name)
            selects[vg.name] = Select(
                self.label, parent=self, name=vg.name, batoms=self.batoms
            )
        return selects

    def __getitem__(self, key):
        """Return a subset of the Batom.

        key -- str, describing which batom to return.
        """
        if isinstance(key, str):
            if key not in self.selects:
                raise KeyError("%s is not in this structure" % key)
            return self.selects[key]
        elif isinstance(key, list):
            raise KeyError("dict not supported yet!")

    def __len__(self):
        return len(self.selects)

    def __repr__(self) -> str:
        s = "     Name        model_style    show  \n"
        i = 0
        for name, sel in self.selects.items():
            s += "{:3d}   {:10s}  \n".format(i, name)
            i += 1
        return s

    def add(self, name, expre=None):
        if expre is not None:
            indices = elect_expression(expre, self.batoms)
        else:
            indices = None
        if len(name) > 4:
            print(
                "Warning: the name of the select: {}, longer than four characters. Has been renamed to {}".format(
                    name, name[:4]
                )
            )
            logger.warning(
                "The name of the select: {}, longer than four characters. Has been renamed to {}".format(
                    name, name[:4]
                )
            )
            name = name[:4]
        # select = self.batoms.attributes['select']
        # select[indices] = string2Number(name)
        # select = {'select': select}
        # self.batoms.set_attributes(select)
        sel = None
        subset = self.find(name)
        if subset is None:
            subset = self.bpy_setting.add()
            self.ui_list_index = len(self.bpy_setting) - 1
            subset.name = name
        subset.label = self.label
        subset.flag = True
        # species_props = self.batoms.
        sel = Select(
            self.label, parent=self, name=name, batoms=self.batoms, indices=indices
        )
        return sel


def elect_expression(expre, batoms):
    """
        select expression
    all
    none

    Logical
    not Sele
    sele1 and sele2

    Properties
    element O

    Coordinates
    z<5

    Chemcial classses

    solvent: water
    hydrogens:
    metals
    """
    indices = None
    if isinstance(expre, (list, np.ndarray)):
        return expre
    if isinstance(expre, str):
        if "chain" in expre:
            chainid = expre.split()[1]
            logger.debug("chain %s" % chainid)
            indices = np.where(batoms.attributes["chainids"] == chainid)[0]
        if "sheet" in expre:
            sheet = expre.split()[1]
            logger.debug("sheet %s" % sheet)
            indices = batoms.ribbon.protein.sheets[sheet].indices
        if "helix" in expre:
            helix = expre.split()[1]
            logger.debug("helix %s" % helix)
            indices = batoms.ribbon.protein.helixs[helix].indices
        if "hetatm" in expre:
            logger.debug("hetatm")
            indices = np.where(batoms.attributes["types"] == "HETATM")[0]
        if "species" in expre:
            species = expre.split()[1]
            logger.debug("species %s" % species)
            indices = np.where(batoms.attributes["species"] == species)[0]
    return indices



"""This adds a menu to shift-a add object menu.
please look at other menu classes in \\scripts\\startup\\bl_ui
i.e. space_view3d.py
"""
from bpy.types import Menu


def menu_func(self, context):
    self.layout.menu("VIEW3D_MT_batoms_add", icon="MESH_UVSPHERE")
    self.layout.menu("VIEW3D_MT_surface_add", icon="MOD_LATTICE")
    self.layout.menu("VIEW3D_MT_nanotube_add", icon="META_CAPSULE")
    self.layout.menu("VIEW3D_MT_nanoparticle_add", icon="MESH_ICOSPHERE")


class VIEW3D_MT_batoms_add(Menu):
    bl_idname = "VIEW3D_MT_batoms_add"
    bl_label = "Batoms"
    bl_icon = "batoms_molecule"

    def draw(self, _context):
        layout = self.layout

        layout.operator_context = "INVOKE_REGION_WIN"

        layout.operator("batoms.atoms_add", text="Atoms", icon="LAYER_ACTIVE")
        layout.operator(
            "batoms.molecule_add", text="Molecule", icon="FORCE_LENNARDJONES"
        )
        layout.operator("batoms.bulk_add", text="Bulk", icon="MESH_CUBE")
        layout.operator("batoms.surface_add", text="Surface", icon="VIEW_ORTHO")
        layout.operator(
            "batoms.root_surface_add", text="Root Surface", icon="MOD_LATTICE"
        )
        layout.operator("batoms.smiles_add", text="Smiles", icon="MOD_LATTICE")

        layout.separator()


class VIEW3D_MT_surface_add(Menu):
    bl_idname = "VIEW3D_MT_surface_add"
    bl_label = "Surface"
    # bl_icon = ""

    def draw(self, _context):
        layout = self.layout

        layout.operator_context = "INVOKE_REGION_WIN"

        layout.operator("surface.fcc100_add", text="fcc100", icon="MESH_PLANE")
        layout.operator("surface.fcc110_add", text="fcc110", icon="MESH_PLANE")
        layout.operator("surface.fcc111_add", text="fcc111", icon="MESH_PLANE")
        layout.operator("surface.fcc211_add", text="fcc211", icon="MESH_PLANE")
        layout.operator(
            "surface.fcc111_root_add", text="fcc111_root", icon="MESH_PLANE"
        )
        layout.separator()
        layout.operator("surface.bcc100_add", text="bcc100", icon="MESH_PLANE")
        layout.operator("surface.bcc110_add", text="bcc110", icon="MESH_PLANE")
        layout.operator("surface.bcc111_add", text="bcc111", icon="MESH_PLANE")
        layout.operator(
            "surface.bcc111_root_add", text="bcc111_root", icon="MESH_PLANE"
        )
        layout.separator()
        layout.operator("surface.hcp0001_add", text="hcp0001", icon="MESH_PLANE")
        layout.operator("surface.hcp10m10_add", text="hcp10m10", icon="MESH_PLANE")
        layout.operator(
            "surface.hcp0001_root_add", text="hcp0001_root", icon="MESH_PLANE"
        )
        layout.separator()
        layout.operator("surface.diamond100_add", text="diamond100", icon="MESH_PLANE")
        layout.operator("surface.diamond111_add", text="diamond111", icon="MESH_PLANE")


class VIEW3D_MT_nanotube_add(Menu):
    bl_idname = "VIEW3D_MT_nanotube_add"
    bl_label = "Nanotube"
    # bl_icon = ""

    def draw(self, _context):
        layout = self.layout

        layout.operator_context = "INVOKE_REGION_WIN"

        layout.operator("nano.nanotube_add", text="Nanotube", icon="MESH_CYLINDER")
        layout.operator("nano.nanoribbon_add", text="nanoribbon", icon="MESH_PLANE")


class VIEW3D_MT_nanoparticle_add(Menu):
    bl_idname = "VIEW3D_MT_nanoparticle_add"
    bl_label = "Nanoparticle"
    # bl_icon = ""

    def draw(self, _context):
        layout = self.layout

        layout.operator_context = "INVOKE_REGION_WIN"

        layout.operator("nano.decahedron_add", text="Decahedron", icon="MESH_ICOSPHERE")
        layout.operator(
            "nano.icosahedron_add", text="Icosahedron", icon="MESH_ICOSPHERE"
        )
        layout.operator("nano.octahedron_add", text="Octahedron", icon="MESH_ICOSPHERE")



import bpy
from bpy.props import (
    FloatVectorProperty,
)
from batoms import Batoms
from batoms.ops.base import OperatorBatoms


class RenderAdd(OperatorBatoms):
    bl_idname = "batoms.render_add"
    bl_label = "Attach Render"
    bl_description = "Add a render to a Batoms object."

    viewport: FloatVectorProperty(
        name="viewport",
        default=(0, 0, 1),
        size=3,
        soft_min=-1,
        soft_max=1,
        description="Viewport",
    )

    resolution: FloatVectorProperty(
        name="resolution",
        default=(1000, 1000),
        size=2,
        soft_min=-1,
        soft_max=1,
        description="resolution",
    )

    def execute(self, context):
        from batoms.render.render import Render

        batoms = Batoms(label=context.object.batoms.label)
        render = Render(
            viewport=self.viewport,
            resolution=self.resolution,
        )
        batoms.render = render
        batoms.render.init()
        bpy.context.space_data.region_3d.view_perspective = "CAMERA"
        return {"FINISHED"}



"""
#TODO wrap is 3d vector, do we need support all?
"""
import bpy
from bpy.types import Panel
from bpy.props import (
    BoolProperty,
    FloatProperty,
    EnumProperty,
    StringProperty,
)
from batoms import Batoms
from batoms.gui.utils import (
    get_active_bpy_data,
    get_attr,
    get_enum_attr,
    set_attr,
    set_enum_attr,
)

model_style_items = [
    ("Space-filling", "Space-filling", "", 0),
    ("Ball-and-stick", "Ball-and-Stick", "", 1),
    ("Polyhedral", "Polyhedral", "", 2),
    ("Stick", "Stick", "", 3),
]


class Batoms_PT_prepare(Panel):
    bl_label = "Batoms"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_options = {"DEFAULT_CLOSED"}
    bl_category = "Batoms"
    bl_idname = "BATOMS_PT_Tools"

    def draw(self, context):
        name = "None"
        if context.object:
            if context.object.batoms.type != "OTHER":
                name = context.object.batoms.label
        layout = self.layout
        layout.label(text="Active: " + name)
        batoms = context.scene.batoms.batoms

        layout.label(text="Model style")
        col = layout.column()
        col.prop(batoms, "model_style", expand=True)
        # layout.label(text="Add label")
        layout.label(text="Radius style")
        layout.prop(batoms, "radius_style", expand=True)
        layout.label(text="Color style")
        layout.prop(batoms, "color_style", expand=True)
        layout.label(text="Polyhedra style")
        layout.prop(batoms, "polyhedra_style", expand=True)

        split = layout.split()
        col = split.column()
        col.prop(batoms, "show", expand=True)
        col = split.column()
        col.prop(batoms, "wrap", expand=True)
        layout.prop(batoms, "crystal_view", expand=True)
        layout.prop(batoms, "show_label", expand=True, text="label")
        layout.prop(batoms, "scale")

        # layout.operator("batoms.replace")


# ---------------------------------------------------
def get_active_batoms():
    context = bpy.context
    if context.object and context.object.batoms.type != "OTHER":
        mode = context.object.mode
        batoms = Batoms(label=context.object.batoms.label)
        bpy.context.view_layer.objects.active = batoms.obj
        bpy.ops.object.mode_set(mode=mode)
        return batoms
    return None


def set_batoms_attr(key, value):
    """ """
    batoms = get_active_batoms()
    if batoms is not None:
        setattr(batoms, key, value)
        bpy.context.view_layer.objects.active = batoms.obj


def modify_batoms_attr(context, key, value):
    """ """
    if context.object and context.object.batoms.type != "OTHER":
        batoms = Batoms(label=context.object.batoms.label)
        setattr(batoms, key, value)
        # batoms.obj.select_set(True)
        bpy.context.view_layer.objects.active = batoms.obj


def get_wrap(self):
    batoms = get_active_bpy_data("batoms")()
    if batoms is not None:
        return batoms.wrap[0]
    else:
        return 0


class BatomsProperties(bpy.types.PropertyGroup):
    model_style: EnumProperty(
        name="model_style",
        description="Structural models",
        items=model_style_items,
        get=get_enum_attr("model_style", get_active_bpy_data("batoms")),
        set=set_enum_attr("model_style", set_batoms_attr),
        default=0,
    )

    show_label: StringProperty(
        name="label",
        description="Show label: None, Index, Species or Charge and so on",
        get=get_attr("show_label", get_active_bpy_data("batoms")),
        set=set_attr("show_label", set_batoms_attr),
        default="",
    )

    radius_style: EnumProperty(
        name="radius_style",
        description="Structural models",
        items=(
            ("Covalent", "covalent", "", 0),
            ("VDW", "van der Waals", "", 1),
            ("Ionic", "ionic", "", 2),
        ),
        get=get_enum_attr("radius_style", get_active_bpy_data("batoms")),
        set=set_enum_attr("radius_style", set_batoms_attr),
        default=0,
    )

    color_style: EnumProperty(
        name="color_style",
        description="Color",
        items=(
            ("JMOL", "JMOL", "", 0),
            ("VESTA", "VESTA", "", 1),
            ("CPK", "CPK", "", 2),
        ),
        get=get_enum_attr("color_style", get_active_bpy_data("batoms")),
        set=set_enum_attr("color_style", set_batoms_attr),
        default=0,
    )

    polyhedra_style: EnumProperty(
        name="polyhedra_style",
        description="Polhhedra models",
        items=(
            ("0", "atoms, bonds and polyhedra", "", 0),
            ("1", "atoms, polyhedra", "", 1),
            ("2", "central atoms, polyhedra", "", 2),
            ("3", "polyhedra", "", 3),
        ),
        get=get_enum_attr("polyhedra_style", get_active_bpy_data("batoms")),
        set=set_enum_attr("polyhedra_style", set_batoms_attr),
        default=0,
    )

    show: BoolProperty(
        name="show",
        default=False,
        description="show all object for view and rendering",
        get=get_attr("show", get_active_bpy_data("batoms")),
        set=set_attr("show", set_batoms_attr),
    )

    wrap: BoolProperty(
        name="wrap",
        default=False,
        description="wrap all atoms into cell",
        get=get_wrap,
        set=set_attr("wrap", set_batoms_attr),
    )

    crystal_view: BoolProperty(
        name="crystal_view",
        default=False,
        description="crystal_view all object for view and rendering",
        get=get_attr("crystal_view", get_active_bpy_data("batoms")),
        set=set_attr("crystal_view", set_batoms_attr),
    )

    scale: FloatProperty(
        name="scale",
        default=1.0,
        min=0.0,
        soft_max=2.0,
        description="scale",
        get=get_attr("scale", get_active_bpy_data("batoms")),
        set=set_attr("scale", set_batoms_attr),
    )



from .cavity import Cavity

from . import (
    bpy_data,
    ops,
    gui,
)

__all__ = ["Cavity"]

classes_bpy_data = [
    # internal data first
    bpy_data.CavitySetting,
    bpy_data.Cavity,
    gui.CavityProperties,
]

classes = [
    ops.CavityAdd,
    ops.CavityRemove,
    ops.CavityDraw,
    gui.VIEW3D_PT_Batoms_cavity,
    gui.BATOMS_MT_cavity_context_menu,
    gui.BATOMS_UL_cavity,
    gui.BATOMS_PT_cavity,
]


def register_class():
    from bpy.types import Collection, Object, Scene
    from bpy.props import PointerProperty
    from bpy.utils import register_class

    for cls in classes_bpy_data:
        register_class(cls)
    # attach to blender internal data
    Collection.Bcavity = PointerProperty(name="Bcavity", type=bpy_data.Cavity)
    Object.Bcavity = PointerProperty(name="Bcavity", type=bpy_data.Cavity)
    Scene.Bcavity = PointerProperty(type=gui.CavityProperties)
    #
    for cls in classes:
        register_class(cls)


def unregister_class():
    from bpy.types import Collection, Object, Scene
    from bpy.utils import unregister_class

    del Collection.Bcavity
    del Object.Bcavity
    del Scene.Bcavity
    for cls in reversed(classes):
        unregister_class(cls)
    for cls in reversed(classes_bpy_data):
        unregister_class(cls)



"""Definition of the Magres class.

This module defines the Magres object in the Batoms package.

"""

import bpy
from time import time
import numpy as np
from batoms.base.object import BaseObject
from .setting import MagresSettings
import logging

logger = logging.getLogger(__name__)


class Magres(BaseObject):
    def __init__(
        self,
        label=None,
        location=np.array([0, 0, 0]),
        batoms=None,
        tensors=None,
    ):
        """Magres Class

        Args:
            label (_type_, optional): _description_. Defaults to None.
            location (_type_, optional): _description_. Defaults to np.array([0, 0, 0]).
            batoms (_type_, optional): _description_. Defaults to None.
        """
        #
        self.batoms = batoms
        self.label = label
        self.tensors = tensors
        name = "magres"
        BaseObject.__init__(self, label, name)
        self.settings = MagresSettings(self.label, parent=self)
        self.settings.bpy_data.active = True

    def build_materials(self, name, color, node_inputs=None, material_style="default"):
        """ """
        from batoms.material import create_material

        if name in bpy.data.materials:
            mat = bpy.data.materials.get(name)
            bpy.data.materials.remove(mat, do_unlink=True)
        mat = create_material(
            name,
            color=color,
            node_inputs=node_inputs,
            material_style=material_style,
            backface_culling=False,
        )
        return mat

    def draw(self, magres_name="ALL"):
        from batoms.utils.butils import clean_coll_object_by_type

        # delete old plane
        clean_coll_object_by_type(self.batoms.coll, "MAGRES")
        for magres in self.settings.bpy_setting:
            if magres_name.upper() != "ALL" and magres.name != magres_name:
                continue
            if magres.type == "MS":
                self.draw_MS(magres)
            elif magres.type == "CS":
                self.draw_CS(magres)

    @property
    def ms_objs(self):
        ms_objs = {}
        for magres in self.settings.bpy_setting:
            ms_name = "%s_%s_ms" % (self.label, magres.name)
            obj = bpy.data.objects.get(ms_name)
            ms_objs[magres.name] = obj
        return ms_objs

    @property
    def ms_mat(self):
        return bpy.data.materials.get(self.ms_name)

    @property
    def cs_objs(self):
        cs_objs = {}
        for magres in self.settings.bpy_setting:
            ms_name = "%s_%s_cs" % (self.label, magres.name)
            obj = bpy.data.objects.get(ms_name)
            cs_objs[magres.name] = obj
        return cs_objs

    def build_ellipsoids(self, positions, tensors, scale):
        """
        generate ellipsoid
        """
        from numpy.linalg import eig

        n = len(tensors)
        ellipsoids = []
        for i in range(n):
            tensor = tensors[i]
            # get symmetry part
            sym_tensor = (tensor + np.transpose(tensor)) / 2.0
            evals, evecs = eig(sym_tensor)
            # abs and normlize
            evals = np.abs(evals)
            evals = evals * scale
            # sort
            index = np.argsort(evals)
            evals = evals[index]
            evecs = evecs[:, index]
            # print(evecs)
            # evecs = evecs/np.linalg.norm(evecs, axis = 0)
            # rotation matrix
            rotation_matrix = np.linalg.inv(evecs)
            ellipsoids.append([positions[i], evals, rotation_matrix])
        return ellipsoids

    def draw_ellipsoids(self, name, datas, coll):
        objs = []
        for data in datas:
            # print(data)
            p = np.eye(4)
            bpy.ops.mesh.primitive_uv_sphere_add(location=data[0], scale=data[1])
            obj = bpy.context.view_layer.objects.active
            me = obj.data
            # rotatation
            p[0:3, 0:3] = data[2]
            me.transform(p)
            me.update()
            objs.append(obj)
        objs[0].name = name
        objs[0].data.name = name
        ctx = bpy.context.copy()
        ctx["object"] = ctx["active_object"] = objs[0]
        ctx["selected_objects"] = ctx["selected_editable_objects"] = objs
        # bpy.context.view_layer.objects.active = obj
        # TODO join objects is not working
        # bpy.ops.object.join(ctx)
        bpy.ops.object.shade_smooth()
        return bpy.data.objects[name]

    def draw_MS(self, magres):
        """ """
        scale = magres.scale
        # print('Scale: {:1.3}'.format(scale))
        tstart = time()
        indices = self.batoms.selects[magres.select].indices
        if len(indices) == 0:
            return
        # select atoms
        positions = self.batoms.positions[indices]
        tensors = np.zeros((len(indices), 3, 3))
        for i in range(3):
            for j in range(3):
                tensors[:, i, j] = self.batoms.get_attribute("ms_%d%d" % (i, j))[
                    indices
                ]
        ellipsoids = self.build_ellipsoids(positions, tensors, scale)
        ms_name = "%s_%s_ms" % (self.label, magres.name)
        self.delete_obj(ms_name)
        coll = self.batoms.coll.children["%s_surface" % self.batoms.coll_name]
        obj = self.draw_ellipsoids(
            ms_name,
            datas=ellipsoids,
            coll=coll,
        )
        mat = self.build_materials(
            ms_name,
            color=magres.color,
            material_style=magres.material_style,
        )
        obj.data.materials.append(mat)
        obj.parent = self.batoms.obj
        obj.batoms.type = "MS"
        obj.batoms.label = self.batoms.label
        logger.debug("Draw MS: %s" % (time() - tstart))

    def draw_CS(self, magres, parallel=1):
        """ """
        pass

    @property
    def setting(self):
        from batoms.utils import deprecated

        """setting object."""
        deprecated(
            '"setting" will be deprecated in the furture, please use "settings".'
        )
        return self.settings

    def as_dict(self):
        """ """
        data = {}
        data["settings"] = self.settings.as_dict()
        data.update(self.settings.bpy_data.as_dict())
        return data



import bpy
import pathlib
import os
import logging

# logger = logging.getLogger('batoms')
logger = logging.getLogger(__name__)


batoms_asset_dir = os.path.join(pathlib.Path(__file__).parent.resolve(), "libraries")


def update_asset_preview():
    for coll in bpy.data.collections:
        if coll.asset_data is not None:
            coll.asset_generate_preview()


def save_asset(filepath, assets):
    bpy.data.libraries.write(filepath, assets)


def push_asset(library, file, assets):
    from pathlib import Path
    import shutil

    # save name of assets
    asset_names = [asset.name for asset in assets]
    # save to local path
    localpath = "./tmp.blend"
    save_asset(localpath, assets)
    #
    path = bpy.context.preferences.filepaths.asset_libraries[library].path
    filepath = os.path.join(path, file)
    if not os.path.exists(filepath):
        emptypath = os.path.join(path, "empty.blend")
        shutil.copyfile(emptypath, filepath)
    bpy.ops.wm.open_mainfile(filepath=filepath)
    localpath = Path(localpath)
    for name in asset_names:
        bpy.ops.wm.append(
            filepath=str(localpath / "Collection" / name),
            directory=str(localpath / "Collection"),
            filename=name,
        )
    bpy.ops.wm.save_mainfile(filepath=filepath, compress=True)
    bpy.ops.wm.open_mainfile(filepath="./tmp.blend")


def find_baotms_catalog_uuid(library, catalog_name, directory=None):
    """Find catalog by name.

    Args:
        catalog_name (str): Name of the catalog
        directory (str, optional): _description_. Defaults to None.

    Returns:
        str: uuid of the catalog.
    """
    if directory is None:
        directory = bpy.context.preferences.filepaths.asset_libraries[library].path
    logger.debug(directory)
    asset_cats = os.path.join(directory, "blender_assets.cats.txt")
    with open(asset_cats) as f:
        for line in f.readlines():
            if line.startswith(("#", "VERSION", "\n")):
                continue
            # uuid:catalog_tree:catalog_name'+'\n'
            name = line.rstrip().split(":")[2]
            if catalog_name == name:
                uuid = line.split(":")[0]
                return uuid


def create_asset(batoms, model_style=0, metadata={}):
    """Create a asset using batoms.

    Args:
        batoms (Batoms): _description_
        metadata (dict): _description_
    """

    batoms.model_style = model_style
    batoms.coll.asset_mark()
    batoms.coll.asset_generate_preview()
    batoms.coll.asset_data.author = metadata["author"]
    batoms.coll.asset_data.description = metadata["description"]
    batoms.coll.asset_data.catalog_id = metadata["catalog_id"]
    for tag in metadata["tags"]:
        batoms.coll.asset_data.tags.new(name=tag)

    return batoms.coll



"""
Use ASE's build function
"""

import bpy
from bpy.types import Operator
from bpy.props import (
    StringProperty,
    IntProperty,
    FloatProperty,
    BoolProperty,
)
from ase.build import graphene_nanoribbon
from batoms import Batoms


class BuildNanoribbon(Operator):
    bl_idname = "nano.nanoribbon_add"
    bl_label = "Add Nanoribbon"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Add Nanoribbon"

    label: StringProperty(name="Label", default="graphene", description="Label")

    type: StringProperty(
        name="Type",
        default="zigzag",
        description=(
            "The orientation of the ribbon. Must be either ‘zigzag’ or ‘armchair’."
        ),
    )

    n: IntProperty(
        name="n",
        default=3,
        min=1,
        soft_max=20,
        description=(
            "The width of the nanoribbon. For armchair nanoribbons, this n may be half-integer to repeat by half a cell."
        ),
    )

    m: IntProperty(
        name="m",
        default=6,
        min=1,
        soft_max=20,
        description="The length of the nanoribbon.",
    )

    C_H: FloatProperty(
        name="C_H",
        default=1.09,
        min=0,
        soft_max=15,
        description="Carbon-hydrogen bond length. Default: 1.09 Angstrom.",
    )

    C_C: FloatProperty(
        name="C_C",
        default=1.42,
        min=0,
        soft_max=15,
        description="Carbon-carbon bond length. Default: 1.42 Angstrom.",
    )

    vacuum: FloatProperty(
        name="Vacuum",
        default=5.0,
        min=0,
        soft_max=15,
        description="Amount of vacuum added to non-periodic directions, if present.",
    )

    saturated: BoolProperty(
        name="Saturated",
        default=False,
        description="If true, hydrogen atoms are placed along the edge.",
    )

    magnetic: BoolProperty(
        name="Magnetic", default=False, description="Make the edges magnetic."
    )

    initial_mag: FloatProperty(
        name="initial_mag",
        default=1.2,
        min=0,
        soft_max=15,
        description="Magnitude of magnetic moment if magnetic.",
    )

    sheet: BoolProperty(
        name="Sheet",
        default=False,
        description=(
            "If true, make an infinite sheet instead of a ribbon (default: False)"
        ),
    )

    main_element: StringProperty(
        name="Main_element", default="C", description="Main_element"
    )

    saturate_element: StringProperty(
        name="Saturate_element", default="H", description="saturate_element"
    )

    def execute(self, context):
        atoms = graphene_nanoribbon(
            n=self.n,
            m=self.m,
            type=self.type,
            saturated=self.saturated,
            C_H=self.C_H,
            C_C=self.C_C,
            vacuum=self.vacuum,
            magnetic=self.magnetic,
            initial_mag=self.initial_mag,
            sheet=self.sheet,
            main_element=self.main_element,
            saturate_element=self.saturate_element,
        )
        batoms = Batoms(label=self.label, from_ase=atoms)
        batoms.obj.select_set(True)
        bpy.context.view_layer.objects.active = batoms.obj
        self.report({"INFO"}, "Add nanoribbon {}".format(self.label))
        return {"FINISHED"}



"""
modified from ASE

Module to read and write atoms in PDB file format.

"""

import warnings

import numpy as np

from ase.atoms import Atoms
from ase.cell import Cell
from ase.io.espresso import label_to_symbol


def read_atom_line(line):
    """
        Read atom line from pdb format
    ATOM     25  CA  LYS A   5      48.356  17.146  -2.714  1.00  0.00           C
    """

    line = line.rstrip("\n")
    type_record = line[0:6].strip()
    name = line[12:16].strip()
    altloc = line[16]
    resname = line[17:21]
    chainid = line[21]  # Not used
    seq = line[22:26].split()
    if len(seq) == 0:
        resseq = 1
    else:
        resseq = int(seq[0])  # sequence identifier
    # icode = line[26]          # insertion code, not used
    # atomic coordinates
    try:
        coord = np.array(
            [float(line[30:38]), float(line[38:46]), float(line[46:54])],
            dtype=np.float64,
        )
    except ValueError:
        raise ValueError("Invalid or missing coordinate(s)")
    # occupancy & B factor
    try:
        occupancy = float(line[54:60])
    except ValueError:
        occupancy = None  # Rather than arbitrary zero or one
    if occupancy is not None and occupancy < 0:
        warnings.warn("Negative occupancy in one or more atoms")
    try:
        bfactor = float(line[60:66])
    except ValueError:
        # The PDB use a default of zero if the data is missing
        bfactor = 0.0
    # segid = line[72:76] # not used
    symbol = line[76:78].strip().upper()
    return (
        symbol,
        name,
        altloc,
        resname,
        coord,
        occupancy,
        bfactor,
        resseq,
        chainid,
        type_record,
    )


def read_line_cyrstal(line):
    cellpar = [
        float(line[6:15]),  # a
        float(line[15:24]),  # b
        float(line[24:33]),  # c
        float(line[33:40]),  # alpha
        float(line[40:47]),  # beta
        float(line[47:54]),
    ]  # gamma
    cell = Cell.new(cellpar)
    pbc = True
    return cell, pbc


def read_line_sheet(line):
    """
    SHEET    1   A 9 PHE A   6  TRP A  12  0"""
    # Chain identifier
    # sheetId = int(line[6:10].strip())
    # chainId = line[11:14].strip()
    startChain = line[21]
    # Residue sequence number
    startResi = int(line[22:26])
    endChain = line[32]
    endResi = int(line[33:37])
    sheet = {
        "name": "%s-%s-%s-%s" % (startChain, startResi, endChain, endResi),
        # 'sheetId': sheetId,
        # 'chainId': chainId,
        "startChain": startChain,
        "startResi": startResi,
        "endChain": endChain,
        "endResi": endResi,
    }
    return sheet


def read_line_helix(line):
    """
    HELIX    3   3 GLY A  724  LEU A  728  5   5
    """
    startChain = line[19]
    startResi = int(line[21:25])
    endChain = line[31]
    endResi = int(line[33:37])
    helix = {
        "name": "%s-%s-%s-%s" % (startChain, startResi, endChain, endResi),
        # 'sheetId': sheetId,
        # 'chainId': chainId,
        "startChain": startChain,
        "startResi": startResi,
        "endChain": endChain,
        "endResi": endResi,
    }
    return helix


def read_pdb(fileobj, index=-1, read_arrays=True):
    """Read PDB files."""
    if isinstance(fileobj, str):
        fileobj = open(fileobj)
    images = []
    orig = np.identity(3)
    trans = np.zeros(3)
    occ = []
    bfactor = []
    types = []
    residuenames = []
    residuenumbers = []
    atomtypes = []
    chainids = []
    sheet = []
    helix = []

    symbols = []
    positions = []
    cell = None
    pbc = None

    def build_atoms():
        atoms = Atoms(symbols=symbols, cell=cell, pbc=pbc, positions=positions)

        if not read_arrays:
            return atoms

        info = {
            "occupancy": occ,
            "bfactor": bfactor,
            "residuenames": residuenames,
            "atomtypes": atomtypes,
            "residuenumbers": residuenumbers,
            "chainids": chainids,
            "types": types,
        }
        for name, array in info.items():
            if len(array) == 0:
                pass
            elif len(array) != len(atoms):
                warnings.warn(
                    "Length of {} array, {}, "
                    "different from number of atoms {}".format(
                        name, len(array), len(atoms)
                    )
                )
            else:
                atoms.set_array(name, np.array(array))
        atoms.info["sheet"] = sheet
        atoms.info["helix"] = helix
        return atoms

    for line in fileobj.readlines():
        if line.startswith("CRYST1"):
            cell, pbc = read_line_cyrstal(line)
        for c in range(3):
            if line.startswith("ORIGX" + "123"[c]):
                orig[c] = [float(line[10:20]), float(line[20:30]), float(line[30:40])]
                trans[c] = float(line[45:55])
        if line.startswith("ATOM") or line.startswith("HETATM"):
            # line_info = symbol, name, altloc, resname, coord, occupancy,
            #             bfactor, resseq
            line_info = read_atom_line(line)

            try:
                symbol = label_to_symbol(line_info[0])
            except (KeyError, IndexError):
                symbol = label_to_symbol(line_info[1])

            position = np.dot(orig, line_info[4]) + trans
            atomtypes.append(line_info[1])
            residuenames.append(line_info[3])
            if line_info[5] is not None:
                occ.append(line_info[5])
            bfactor.append(line_info[6])
            residuenumbers.append(line_info[7])
            chainids.append(line_info[8])
            types.append(line_info[9])

            symbols.append(symbol)
            positions.append(position)

        if line.startswith("SHEET"):
            sheet.append(read_line_sheet(line))
        if line.startswith("HELIX"):
            helix.append(read_line_helix(line))
        if line.startswith("CONECT"):
            pass
        if line.startswith("REMARK"):
            pass
        if line.startswith("HEADER"):
            pass
        if line.startswith("TITLE"):
            pass
        if line.startswith("COMPND"):
            pass
        if line.startswith("END"):
            # End of configuration reached
            # According to the latest PDB file format (v3.30),
            # this line should start with 'ENDMDL' (not 'END'),
            # but in this way PDB trajectories from e.g. CP2K
            # are supported (also VMD supports this format).
            atoms = build_atoms()
            images.append(atoms)
            occ = []
            bfactor = []
            residuenames = []
            atomtypes = []
            symbols = []
            positions = []
            cell = None
            pbc = None

    if len(images) == 0:
        atoms = build_atoms()
        images.append(atoms)
    return images[index]


if __name__ == "__main__":
    images = read_pdb("test/datas/1tim.pdb")
    # images = read_pdb('test/datas/2piw.pdb')
    print(images.arrays.keys())
    print(images.info["helix"])



from bpy.props import (
    StringProperty,
    BoolProperty,
    IntProperty,
    FloatProperty,
    FloatVectorProperty,
    CollectionProperty,
)

from batoms.internal_data import Base


class PolyhedraSetting(Base):
    """ """

    flag: BoolProperty(name="flag", default=False)
    label: StringProperty(name="label", default="")
    species: StringProperty(name="species")
    name: StringProperty(name="name")
    color: FloatVectorProperty(name="color", subtype="COLOR", min=0, max=1, size=4)
    width: FloatProperty(name="width", min=0, soft_max=1, default=0.01)
    show_edge: BoolProperty(name="show_edge", default=True)

    @property
    def name(self) -> str:
        return self.species

    def as_dict(self) -> dict:
        setdict = {
            "flag": self.flag,
            "label": self.label,
            "species": self.species,
            "name": self.name,
            "material_style": self.material_style,
            "color": self.color[:],
            "width": self.width,
            "show_edge": self.show_edge,
        }
        return setdict

    def __repr__(self) -> str:
        s = "-" * 60 + "\n"
        s = "Center                show_edge           width \n"
        s += "{0:10s}    {1:10s}   {2:1.3f} \n".format(
            self.species, str(self.show_edge), self.width
        )
        s += "-" * 60 + "\n"
        return s


class Polyhedra(Base):
    label: StringProperty(name="label", default="batoms")
    # polyhedra_style
    ui_list_index: IntProperty(name="ui_list_index", default=0)
    # collection
    settings: CollectionProperty(name="polyhedrasetting", type=PolyhedraSetting)

    def as_dict(self) -> dict:
        setdict = {}
        return setdict



"""
Logger
"""
import bpy
import sys
from tempfile import gettempdir
from pathlib import Path
import logging
import os

batoms_dir = os.path.dirname(Path(__file__).parent)


# logger = logging.getLogger('batoms')
logger = logging.getLogger(__name__)
root_logger = logging.getLogger("batoms")


def set_logger(version):
    # formatter = ('%(levelname)-8s '
    #                             '[%(funcName)-20s]: %(message)s')
    formatter = "%(levelname)s " "[%(name)-10s %(funcName)-10s]: %(message)s"
    logging.basicConfig(stream=sys.stdout, format=formatter, level=logging.INFO)
    # add logger file
    filepath = Path(gettempdir()) / ("beautiful_atoms.log")
    root_logger.info("Log file: " + str(filepath))
    file_handler = logging.FileHandler(filepath, mode="w")
    file_handler.setFormatter(logging.Formatter(formatter))
    root_logger.addHandler(file_handler)
    root_logger.info("Blender version: {} ".format(bpy.app.version_string))
    root_logger.info("Python version: {} ".format(sys.version))
    root_logger.info("Beautiful Atoms version: {} ".format(version))
    root_logger.info("Beautiful Atoms directory: {} ".format(batoms_dir))


def update_logging_level():
    if "batoms" not in bpy.context.preferences.addons:
        return
    prefs = bpy.context.preferences.addons["batoms"].preferences
    root_logger.setLevel(prefs.logging_level)


def print_time(key, value):
    return "{}: {}".format(key, value)



import bpy
import numpy as np
import pytest


def test_settings(ch4):
    """species panel"""
    assert ch4.coll.batoms.ui_list_index_species == 1
    # add
    ch4.species.add("H_1")
    assert ch4.coll.batoms.ui_list_index_species == 2
    # remove
    ch4.species.remove("H_1")
    assert ch4.coll.batoms.ui_list_index_species == 1
    # ops add
    bpy.context.view_layer.objects.active = ch4.obj
    bpy.ops.batoms.species_add(species="H_1")
    assert ch4.coll.batoms.ui_list_index_species == 2
    # ops remove
    bpy.ops.batoms.species_remove(species="H_1")
    assert ch4.coll.batoms.ui_list_index_species == 1


def test_batoms_species(h2o):
    """Setting properties of species"""
    assert len(h2o.species) == 2
    # default covalent radius
    assert np.isclose(h2o.radius["H"], 0.31)
    # vdw radius
    h2o.radius_style = 1
    assert np.isclose(h2o.radius["H"], 1.2)
    # default Jmol color
    assert np.isclose(h2o["H"].color, np.array([1, 1, 1, 1])).all()
    # VESTA color
    h2o.color_style = 2
    assert np.isclose(h2o["H"].color, np.array([1, 0.8, 0.8, 1])).all()
    # materials
    h2o.species["H"].materials = {"Metallic": 0.9, "IOR": 1.0, "Roughness": 0.01}
    # elements
    h2o.species["H"].elements = {"H": 0.5, "O": 0.4}
    assert len(h2o["H"].elements) == 3
    h2o["H"].material_style = "mirror"
    # species X
    h2o.replace([0], "X")
    h2o["X"].color = [0.8, 0.8, 0.0, 0.3]


def test_species_color(ch4):
    ch4.species["C"].color = [1, 0, 0, 1]
    ch4.species["C"].material_style = "metallic"
    assert np.isclose(ch4.species["C"].color, np.array([1, 0, 0, 1])).all()
    ch4.species.update()
    assert np.isclose(ch4.species["C"].color, np.array([1, 0, 0, 1])).all()


def test_auto_build_species():
    """auto build species use spglib"""
    pytest.importorskip("spglib")
    from batoms.bio.bio import read

    bpy.ops.batoms.delete()
    magnetite = read("../tests/datas/magnetite.cif")
    magnetite.auto_build_species()
    assert len(magnetite.species) == 5
    assert len(magnetite.bonds.setting) == 10
    assert len(magnetite.polyhedras.setting) == 3


def test_geometry_node_object(tio2):
    """species instances are used in geometry node,
    in batoms, boundary, search_bond. When instances are
    re-build, we need also update the geometry node."""
    tio2.boundary = 0.01
    tio2.bond.show_search = True
    tio2.model_style = 1
    tio2.species["Ti"].color = [1, 1, 0, 1]
    tio2.species.update()
    species_node = tio2.gn_node_group.nodes["Atoms_tio2"].node_tree.nodes[
        "Atoms_tio2_Ti"
    ]
    assert species_node.inputs["Instancer"].default_value is not None
    assert (
        tio2.boundary.gn_node_group.nodes["ObjectInfo_tio2_Ti"]
        .inputs["Object"]
        .default_value
        is not None
    )
    assert (
        tio2.bond.search_bond.gn_node_group.nodes["ObjectInfo_tio2_Ti"]
        .inputs["Object"]
        .default_value
        is not None
    )


def test_color_by_attribute():
    from ase.build import bulk
    from batoms import Batoms
    import numpy as np

    au = bulk("Au", cubic=True) * [5, 5, 5]
    au = Batoms(label="au", from_ase=au)
    z = au.positions[:, 2]
    au.set_attributes({"z_coor": (z - np.min(z)) / (np.max(z) - np.min(z))})
    au.species.color_by_attribute("z_coor")
    assert (
        au.species["Au"].materials["Au"].node_tree.nodes["Attribute"].attribute_name
        == "z_coor"  # noqa W503
    )
    assert len(au.species["Au"].materials["Au"].node_tree.links) > 2



import bpy
import numpy as np
from batoms.utils.butils import (
    get_node_by_name,
    object_mode,
    set_look_at,
    update_object,
)

import bmesh
import logging

# logger = logging.getLogger('batoms')
logger = logging.getLogger(__name__)


default_object_attributes = []

default_object_datas = {}


class BaseObject:
    def __init__(self, obj_name, btype="batoms"):
        self.obj_name = obj_name
        self.btype = btype

    @property
    def obj(self):
        return self.get_obj()

    def get_obj(self):
        obj = bpy.data.objects.get(self.obj_name)
        if obj is None:
            raise KeyError("%s object is not exist." % self.obj_name)
        return obj

    @property
    def bpy_data(self):
        return self.get_bpy_data()

    def get_bpy_data(self):
        bpy_data = getattr(self.obj, self.btype)
        return bpy_data

    @property
    def location(self):
        return self.get_location()

    def get_location(self):
        return np.array(self.obj.location)

    @location.setter
    def location(self, location):
        self.set_location(location)

    def set_location(self, location):
        self.obj.location = location

    @property
    def type(self):
        return self.get_type()

    # this is weak, because not work for mesh object

    @type.setter
    def type(self, type):
        self.set_type(type)

    def get_type(self):
        return self.obj.data.type

    def set_type(self, type):
        self.obj.data.type = type.upper()

    @property
    def hide(self):
        return self.get_hide()

    @hide.setter
    def hide(self, state):
        self.set_hide(state)

    def get_hide(self):
        return self.obj.hide_get()

    def set_hide(self, state):
        self.obj.hide_render = state
        self.obj.hide_set(state)

    @property
    def select(self):
        return self.get_select()

    @select.setter
    def select(self, state):
        self.set_select(state)

    def get_select(self):
        return self.obj.select_get()

    def set_select(self, state):
        self.obj.select_set(state)

    @property
    def scene(self):
        return self.get_scene()

    def get_scene(self):
        return bpy.context.scene

    @property
    def look_at(self):
        return self.get_look_at()

    @look_at.setter
    def look_at(self, look_at):
        self.set_look_at(look_at)

    def get_look_at(self):
        return np.array(self.bpy_data.look_at)

    def set_look_at(self, look_at):
        self.bpy_data.look_at = look_at
        set_look_at(self.obj, look_at, roll=0.0)

    def translate(self, displacement):
        """Translate atomic positions.

        The displacement argument is an xyz vector.

        For example, move H species molecule by a vector [0, 0, 5]

        >>> h.translate([0, 0, 5])
        """
        object_mode()
        bpy.ops.object.select_all(action="DESELECT")
        self.obj.select_set(True)
        bpy.ops.transform.translate(value=displacement)

    def rotate(self, angle, axis="Z", orient_type="GLOBAL"):
        """Rotate atomic based on a axis and an angle.

        Parameters:

        angle: float
            Angle that the atoms is rotated around the axis.
        axis: str
            'X', 'Y' or 'Z'.

        For example, rotate h2o molecule 90 degree around 'Z' axis:

        >>> h.rotate(90, 'Z')

        """
        object_mode()
        bpy.ops.object.select_all(action="DESELECT")
        self.obj.select_set(True)
        bpy.context.view_layer.objects.active = self.obj
        bpy.ops.transform.rotate(
            value=angle, orient_axis=axis.upper(), orient_type=orient_type
        )

    def delete_obj(self, name):
        if name in bpy.data.objects:
            obj = bpy.data.objects.get(name)
            bpy.data.objects.remove(obj, do_unlink=True)

    def delete_material(self, name):
        if name in bpy.data.materials:
            obj = bpy.data.materials.get(name)
            bpy.data.materials.remove(obj, do_unlink=True)

    def update_mesh(self, obj=None):
        """Update mesh of the object.
        By switching to edit mode and back to object mode."""

        object_mode()
        if obj is None:
            obj = self.obj
        # bm = bmesh.new()
        # bm.from_mesh(obj.data)
        # bm.verts.ensure_lookup_table()
        # bm.to_mesh(obj.data)
        # bm.clear()
        # bpy.context.view_layer.update()
        # I don't why the shape key is not udpate in background mode, so we need this.
        bpy.context.view_layer.objects.active = self.obj
        mode = self.obj.mode
        bpy.ops.object.mode_set(mode="EDIT")
        bpy.ops.object.mode_set(mode="OBJECT")
        bpy.ops.object.mode_set(mode=mode)

    def add_verts(self, count, obj=None):
        """Add vertices to the object."""

        object_mode()
        if obj is None:
            obj = self.obj
        obj.data.vertices.add(count)
        self.update_mesh(obj)

    def add_vertices_bmesh(self, count, obj=None):
        import bmesh
        import numpy as np

        object_mode()
        if obj is None:
            obj = self.obj
        bm = bmesh.new()
        bm.from_mesh(obj.data)
        bm.verts.ensure_lookup_table()
        vertices = np.zeros((count, 3))
        for vert in vertices:
            bm.verts.new(vert)
        bm.to_mesh(obj.data)
        bm.clear()

    def delete_vertices_bmesh(
        self,
        index=[],
        obj=None,
    ):
        """
        delete verts
        """
        import bmesh

        object_mode()
        if obj is None:
            obj = self.obj
        bm = bmesh.new()
        bm.from_mesh(obj.data)
        bm.verts.ensure_lookup_table()
        verts_select = [bm.verts[i] for i in index]
        bmesh.ops.delete(bm, geom=verts_select, context="VERTS")
        bm.to_mesh(obj.data)
        bm.clear()

    def add_edges(self, edges=None, obj=None):
        object_mode()
        if obj is None:
            obj = self.obj
        obj.data.edges.add(len(edges))
        edges = edges.reshape(len(edges) * 2)
        obj.data.edges.foreach_set("vertices", edges)
        self.update_mesh(obj)

    def delete_edges_bmesh(
        self,
        index=[],
        obj=None,
    ):
        """
        delete edges
        """
        import bmesh

        object_mode()
        if obj is None:
            obj = self.obj
        bm = bmesh.new()
        bm.from_mesh(obj.data)
        bm.edges.ensure_lookup_table()
        edges_select = [bm.edges[i] for i in index]
        bmesh.ops.delete(bm, geom=edges_select, context="EDGES_FACES")
        bm.to_mesh(obj.data)
        bm.clear()

    @property
    def shape_keys(self):
        base_name = "Basis_%s" % self.obj_name
        if self.obj.data.shape_keys is None and len(self) > 0:
            self.obj.shape_key_add(name=base_name)
        return self.obj.data.shape_keys

    def __len__(self):
        return len(self.obj.data.vertices)


class ObjectGN(BaseObject):
    """
    Object with Geometry Node.

    """

    def __init__(self, label, name=None):
        if name:
            self.name = name
            obj_name = "%s_%s" % (label, name)
        else:
            obj_name = label
        BaseObject.__init__(self, obj_name=obj_name)

    def build_object(self, arrays, attributes={}):
        self.set_attributes(attributes)
        self.build_geometry_node()
        self.set_shape_key(self._trajectory)

    def load(self):
        flag = True
        obj = bpy.data.objects.get(self.obj_name)
        if obj is None:
            flag = False
        return flag

    @property
    def gn_modifier(self):
        return self.get_gn_modifier()

    @property
    def gn_node_group(self):
        """Get the geometry node group of the object."""
        return self.gn_modifier.node_group

    def get_gn_modifier(self):
        """Get the geometry node modifier of the object."""
        name = "GeometryNodes_%s" % self.obj_name
        modifier = self.obj.modifiers.get(name)
        if modifier is None:
            self.init_geometry_node_modifier()
        return modifier

    def init_geometry_node_modifier(self, inputs=[]):
        """Init geometry node modifier"""
        # blender 4.0 use a interface to add sockets, both input and output
        from batoms.utils.butils import build_gn_modifier

        name = "GeometryNodes_%s" % self.obj_name
        modifier = build_gn_modifier(self.obj, name)
        if bpy.app.version_string >= "4.0.0":
            interface = modifier.node_group.interface
            for input in inputs:
                socket = interface.new_socket(
                    name=input[0], socket_type=input[1], in_out="INPUT"
                )
                modifier["%s_use_attribute" % socket.identifier] = True
                modifier["%s_attribute_name" % socket.identifier] = input[0]
        else:
            for input in inputs:
                modifier.node_group.inputs.new(input[1], input[0])
                id = modifier.node_group.inputs[input[0]].identifier
                modifier["%s_use_attribute" % id] = True
                modifier["%s_attribute_name" % id] = input[0]
        return modifier

    def build_geometry_node(self):
        """
        Geometry node for everything!
        """
        from batoms.utils.butils import build_gn_modifier

        name = "GeometryNodes_%s" % self.obj_name
        modifier = build_gn_modifier(self.obj, name)
        return modifier

    def vectorDotMatrix(self, gn_node_group, vector_output, matrix, name):
        """ """
        CombineXYZ = get_node_by_name(
            gn_node_group.nodes,
            "%s_CombineXYZ_%s" % (self.label, name),
            "ShaderNodeCombineXYZ",
        )
        #
        VectorDot = []
        for i in range(3):
            tmp = get_node_by_name(
                gn_node_group.nodes,
                "%s_VectorDot%s_%s" % (self.label, i, name),
                "ShaderNodeVectorMath",
            )
            tmp.operation = "DOT_PRODUCT"
            VectorDot.append(tmp)
            tmp.inputs[1].default_value = matrix[:, i]
            gn_node_group.links.new(vector_output, tmp.inputs[0])
            gn_node_group.links.new(tmp.outputs["Value"], CombineXYZ.inputs[i])
        return CombineXYZ

    def add_geometry_node(self):
        """
        add geometry node for species
        """
        return NotImplementedError("add_geometry_node is not implemented.")

    @property
    def realize_instances(self):
        return self.get_realize_instances()

    @realize_instances.setter
    def realize_instances(self, state):
        self.set_realize_instances(state)

    def get_realize_instances(self):
        return list(self.coll.batoms.realize_instances)

    def set_realize_instances(self, realize_instances):
        """Make instancing object real
        # TODO: add make real to geometry node
        """
        #
        nodes = self.gn_node_group.nodes
        RealizeInstances = get_node_by_name(
            self.gn_node_group.nodes,
            "%s_RealizeInstances" % self.label,
            "GeometryNodeRealizeInstances",
        )
        if not realize_instances:
            # switch off
            if len(RealizeInstances.outputs[0].links) > 0:
                link = RealizeInstances.outputs[0].links[0]
                self.gn_node_group.links.remove(link)
            self.gn_node_group.links.new(
                nodes["%s_JoinGeometry" % self.label].outputs[0], nodes[1].inputs[0]
            )
        else:
            self.gn_node_group.links.new(
                nodes["%s_JoinGeometry" % self.label].outputs[0],
                RealizeInstances.inputs[0],
            )
            self.gn_node_group.links.new(
                RealizeInstances.outputs[0], nodes[1].inputs[0]
            )
        self.gn_node_group.update_tag()

    def update(
        self,
    ):
        """
        update object'data.
        calculate data in all farmes
        """
        pass

    @property
    def location(self):
        return self.get_location()

    def get_location(self):
        return np.array(self.obj.location)

    @location.setter
    def location(self, location):
        self.set_location(location)

    def set_location(self, location):
        self.obj.location = location

    @property
    def arrays(self):
        return self.get_arrays()

    @arrays.setter
    def arrays(self, arrays):
        self.set_arrays(arrays)

    def get_arrays(self):
        """Get the arrays of the Batoms object.

        arrays is the combination of attributes and the  positions


        Returns:
            dict: _description_
        """
        object_mode()
        arrays = self.attributes
        arrays.update(
            {
                "positions": self.positions[0],
            }
        )
        return arrays

    @property
    def attributes(self):
        return self.get_attributes()

    @attributes.setter
    def attributes(self, attributes):
        self.set_attributes(attributes)

    def get_attributes(self, domains=["POINT"]):
        """Get all attributes of the Batoms object.


        Returns:
            dict: attributes dict
        """
        attributes = {}
        for att in self.obj.data.attributes:
            if att.name.startswith(".") or att.domain not in domains:
                continue
            attributes[att.name] = self.get_attribute(att.name)
        return attributes

    def get_attribute(self, key, index=None):
        """Get an attribute of Batoms object by name
        If the attribute is single value, just use the attribute of mesh.
        If the attribute is an array, we should gather several attributes
        from the mesh.

        Args:
            key (string): name of the attribute

        Raises:
            KeyError: _description_

        Returns:
            _type_: _description_
        """
        from batoms.utils.attribute import get_mesh_attribute, get_mesh_attribute_bmesh

        # get the mesh
        obj = self.obj
        att = obj.data.attributes[key]
        if obj.mode == "EDIT" and att.data_type in [
            "STRING",
            "INT",
            "FLOAT",
            "FLOAT_VECTOR",
            "FLOAT_COLOR",
        ]:
            attribute = get_mesh_attribute_bmesh(obj, att.name, index)
        else:
            attribute = get_mesh_attribute(obj, att.name, index)

        return attribute

    def add_attribute_from_array(self, name, data, domain="POINT"):
        """Add an attribute from array"""
        from batoms.utils import type_py_to_blender

        # check the type of the array, and convert it to blender type
        # only the first element is checked
        # check if shape of the data array
        # if 1x2, 1x3, 1x4
        if isinstance(data, list):
            data = np.array(data)
        try:
            if len(data.shape) == 1:
                array = np.asarray(data)
                type_py = type(array.flat[0])
                dtype_bl = type_py_to_blender(type_py)
                if dtype_bl is False:
                    print(
                        f"Attribute: {name} is not added. The type of the array: {type_py} is not supported."  # noqa E501
                    )
                    return False
                self.add_attribute(name=name, data_type=dtype_bl, domain=domain)
            elif len(data.shape) == 2:
                if data.shape[1] == 2:
                    self.add_attribute(name=name, data_type="FLOAT2", domain=domain)
                elif data.shape[1] == 3:
                    self.add_attribute(
                        name=name, data_type="FLOAT_VECTOR", domain=domain
                    )
                elif data.shape[1] == 4:
                    self.add_attribute(name=name, data_type="QUATERNION", domain=domain)
                else:
                    print(
                        f"Attribute: {name} is not added. The shape of the array: {data.shape} is not supported."  # noqa E501
                    )
                    return False
            else:
                # print a Warning message
                print(
                    f"Attribute: {name} is not added. The shape of the array: {data.shape} is not supported."  # noqa E501
                )
                return False
        except Exception:
            print(
                f"Attribute: {name} is not added. The shape of the array: {data.shape} is not supported."  # noqa E501
            )
            return False

    def add_attribute(self, name, data_type="FLOAT", domain="POINT"):
        """Add an attribute to the mesh"""
        self.obj.data.attributes.new(name=name, type=data_type, domain=domain)
        return True

    def set_attributes(self, attributes):
        """Set attributes

        Args:
            attributes (dict): attributes bound to every atoms
        """
        obj = self.obj
        me = obj.data
        # print(attributes)
        for key, array in attributes.items():
            # if key in self._attributes
            if key not in obj.data.attributes:
                # try to create a new attribute
                flag = self.add_attribute_from_array(key, array)
                logger.debug(f"Add new attribute: {key}")
                # if failed, do not add this attribute
                if flag is False:
                    continue
            self.set_attribute(key, array)
        me.update()

    def set_attribute(self, key, array, index=None):
        """Set one attribute

        An attribute is a generic term to describe data stored per-element
        in a geometry data-block. For example, every vertex can have an
        associated number or vector.

        class bpy.types.Attribute

        Two cases:
        1) for single value data (dimension=0), treat it as normal
        2) for array data (dimension > 0), scatter the data.

        Two mode:
        1) Edit mode, use bmesh
        2) Object mode, use data.attributes directly

        Special case:
        1) String, can not use foreach_set. Must use bmesh with encode,
            otherwise, can not read use bmesh.
        2) Boolean, does not supported by bmesh. Must use Object mode.
           We set all Boolean properties to INT.

        Args:
            key (str): name of the attribute
            array (np.array): value of the attribute
        """
        from batoms.utils.attribute import set_mesh_attribute, set_mesh_attribute_bmesh

        obj = self.obj
        me = obj.data
        if key not in me.attributes:
            raise KeyError(
                "Attribute: {} is not exist. Please add it first.".format(key)
            )
        array = np.array(array)
        if len(array.shape) == 0:
            array = np.array([array])
        if len(array) == 0:
            return
        # single value data
        att = me.attributes.get(key)
        if att.data_type == "STRING" or (
            obj.mode == "EDIT" and att.data_type in ["INT", "FLOAT"]
        ):
            set_mesh_attribute_bmesh(obj, key, array, index)
        else:
            set_mesh_attribute(obj, key, array, index)

    def set_attribute_with_indices(self, name, indices, data):
        data0 = self.get_attribute(name)
        data0[indices] = data
        self.set_attributes({name: data0})

    def get_attribute_with_indices(self, name, indices):
        """Get attribute with indices, with len(indices) > 1

        Read the whole attribute, and then slice.

        If len(indices) == 1, use get_attribute(key, index=indices[0]),
        this is faster.

        Args:
            name (str): _description_
            indices (array): _description_

        Returns:
            _type_: _description_
        """
        return self.get_attribute(name)[indices]

    @property
    def positions(self):
        return self.get_positions()

    def get_positions(self):
        """Get the local positions of the object."""
        n = len(self)
        positions = np.empty(n * 3, dtype=np.float64)
        self.obj.data.vertices.foreach_get("co", positions)
        positions = positions.reshape((n, 3))
        return positions

    @positions.setter
    def positions(self, positions):
        self.set_positions(positions)

    def set_positions(self, positions):
        """
        Set positions to local vertices
        """
        object_mode()
        natom = len(self)
        n = len(positions)
        if n != natom:
            raise ValueError("positions has wrong shape %s != %s." % (n, natom))
        if natom == 0:
            return
        positions = positions.reshape((natom * 3, 1))
        self.obj.data.vertices.foreach_set("co", positions)
        self.obj.data.update()
        bpy.context.view_layer.objects.active = self.obj
        bpy.ops.object.mode_set(mode="EDIT")
        bpy.ops.object.mode_set(mode="OBJECT")

    @property
    def global_positions(self):
        return self.get_global_positions()

    @global_positions.setter
    def global_positions(self, global_positions):
        self.set_global_positions(global_positions)

    def get_global_positions(self):
        """
        Get global positions.
        """
        from batoms.utils import local2global

        global_positions = local2global(self.positions, np.array(self.obj.matrix_world))
        return global_positions

    def set_global_positions(self, positions):
        """
        Set global positions to local vertices
        """
        object_mode()
        from batoms.utils import local2global

        natom = len(self)
        nposition = len(positions)
        if nposition != natom:
            raise ValueError("positions has wrong shape %s != %s." % (nposition, natom))
        if natom == 0:
            return
        local_positions = local2global(
            positions, np.array(self.obj.matrix_world), reversed=True
        )
        self.local_positions = local_positions

    @property
    def nframe(self):
        return self.get_nframe()

    def get_nframe(self):
        if self.obj.data.shape_keys is None:
            return 0
        nframe = len(self.obj.data.shape_keys.key_blocks)
        return nframe

    def get_shape_key(self, obj, local=True):
        """
        read shape key
        """
        from batoms.utils import local2global

        n = len(self)
        nframe = self.nframe
        trajectory = np.empty((nframe, n, 3), dtype=np.float64)
        for i in range(nframe):
            positions = np.empty(n * 3, dtype=np.float64)
            sk = obj.data.shape_keys.key_blocks[i]
            sk.data.foreach_get("co", positions)
            local_positions = positions.reshape((n, 3))
            if local:
                trajectory[i] = local_positions
            else:
                global_positions = local2global(
                    local_positions, np.array(self.obj.matrix_world)
                )
                trajectory[i] = global_positions
        return trajectory

    def set_shape_key(self, name, obj, trajectory=None, frame_start=0):
        """Set shape key for trajectory

        name: str
            name of the shape key
        obj: bpy.data.objects
            object to set shape key
        trajectory: list
            list of positions
        frame_start: int
            start frame
        """
        from batoms.utils.butils import add_keyframe_to_shape_key

        if trajectory is None:
            trajectory = self.trajectory
        nframe = len(trajectory)
        if nframe == 0:
            return
        # shape key should not be add for empty mesh
        if len(trajectory[0]) == 0:
            return
        # name = '%s_bond%s'%(self.label, sp)
        # obj = bpy.data.objects.get(name)
        base_name = "Basis_%s" % (name)
        if obj.data.shape_keys is None:
            sk = obj.shape_key_add(name=base_name)
        elif base_name not in obj.data.shape_keys.key_blocks:
            sk = obj.shape_key_add(name=base_name)
        else:
            sk = obj.data.shape_keys.key_blocks.get(base_name)
        # set basis key
        nvert = len(obj.data.shape_keys.key_blocks[0].data)
        positions = trajectory[0]
        vertices = positions.reshape(nvert * 3)
        sk.data.foreach_set("co", vertices)
        # self.obj.data.update()
        for i in range(1, nframe):
            name = str(i)
            if name not in obj.data.shape_keys.key_blocks:
                sk = obj.shape_key_add(name=name)
                # Add Keytrajectory, the last one is different
                if i != nframe - 1:
                    add_keyframe_to_shape_key(
                        sk,
                        "value",
                        [0, 1, 0],
                        [frame_start + i - 1, frame_start + i, frame_start + i + 1],
                    )
                else:
                    add_keyframe_to_shape_key(
                        sk, "value", [0, 1], [frame_start + i - 1, frame_start + i]
                    )
            else:
                sk = obj.data.shape_keys.key_blocks.get(name)
            # Use the local position here
            positions = trajectory[i]
            vertices = positions.reshape((nvert * 3, 1))
            sk.data.foreach_set("co", vertices)
        self.update_mesh(obj)

    def delete_obj(self, name):
        if name in bpy.data.objects:
            obj = bpy.data.objects.get(name)
            # if not obj.batoms.flag:
            # raise Exception(
            # "Failed, the name {} already in use and is not
            # Batom object!".format(name))
            bpy.data.objects.remove(obj, do_unlink=True)


class childObjectGN:
    """
    Slice of Object with Geometry Node.

    """

    def __init__(self, label, indices, obj_name=None, parent=None):
        self.label = label
        self.parent = parent
        if isinstance(indices, slice):
            self.indices = np.arange(len(self.parent))[indices]
        elif isinstance(indices, int):
            self.indices = np.array([indices])
        else:
            self.indices = np.array(indices)

        if obj_name is None:
            self.obj_name = label
        else:
            self.obj_name = obj_name

    @property
    def obj(self):
        return self.get_obj()

    def get_obj(self):
        obj = bpy.data.objects.get(self.obj_name)
        if obj is None:
            raise KeyError("%s object is not exist." % self.obj_name)
        return obj

    @property
    def vertice(self):
        if self.obj.data.shape_keys is None:
            return self.obj.data.vertices[self.indices[0]]
        else:
            return self.obj.data.shape_keys.key_blocks[0].data[self.indices[0]]

    @property
    def bm(self):
        if self.obj.mode == "EDIT":
            bm = bmesh.from_edit_mesh(self.obj.data)
        else:
            bm = bmesh.new()
            bm.from_mesh(self.obj.data)
        return bm

    @property
    def attributes(self):
        return self.get_attributes()

    def get_attributes(self):
        """Get all attributes of the Batoms object.

        Returns:
            dict: attributes dict
        """
        attributes = {}
        for att in self.parent._attributes:
            attributes[att.name] = self.get_attribute(att.name)
        return attributes

    @property
    def local_position(self):
        if len(self.indices) == 1:
            return self.vertice.co
        else:
            return self.parent.local_positions[self.indices]

    @property
    def position(self):
        if len(self.indices) == 1:
            return self.get_position()
        else:
            return self.parent.positions[self.indices]

    @position.setter
    def position(self, value):
        if len(self.indices) == 1:
            self.set_position(value)
        else:
            positions = self.parent.positions
            positions[self.indices] = value
            self.parent.positions[self.indices] = positions

    def get_position(self):
        """
        Get global position.
        """
        from batoms.utils import local2global

        position = local2global(
            np.array([self.local_position]), np.array(self.obj.matrix_world)
        )
        return position[0]

    def set_position(self, value):
        """
        Set global position to local vertices
        """
        object_mode()
        from batoms.utils import local2global

        position = np.array([value])
        position = local2global(
            position, np.array(self.obj.matrix_world), reversed=True
        )
        # rashpe to (natoms*3, 1) and use forseach_set
        self.vertice.co = position[0]
        self.obj.data.update()
        update_object(self.obj)

    @property
    def scale(self):
        return self.get_attribute("scale")

    @scale.setter
    def scale(self, value):
        self.set_attribute("scale", value)

    @property
    def show(self):
        return self.get_attribute("show")

    @show.setter
    def show(self, value):
        self.set_attribute("show", value)

    def get_attribute(self, key):
        """Helper function to get attribute

        Args:
            key (str): name of the attribute

        Returns:
            _type_: _description_
        """
        if len(self.indices) == 1:
            value = self.parent.get_attribute(key, index=self.indices[0])
        else:
            value = self.parent.get_attribute_with_indices(key, self.indices)
        return value

    def set_attribute(self, key, value):
        """Helper function to set attribute

        Args:
            key (str): _description_
            value (_type_): _description_
        """
        if len(self.indices) == 1:
            self.parent.set_attribute(key, value, index=self.indices[0])
        else:
            self.parent.set_attribute_with_indices(key, self.indices, value)



def test_slicebonds(ch4):
    """Setting sliced Bonds"""
    import numpy as np

    ch4.model_style = 1
    # one atoms
    assert ch4.bond[0].order == 1
    ch4.bond[1].order = 2
    assert ch4.bond[1].order == 2
    ch4.bond[1].show = 0
    assert ch4.bond[1].show == 0
    # > one bonds
    assert np.isclose(ch4.bond[0:2].order, np.array([1, 2])).all()
    ch4.bond[0:2].order = 2
    assert np.isclose(ch4.bond[0:2].order, np.array([2, 2])).all()



import bpy
from batoms.render import Render, Lights
from batoms.build import molecule

try:
    from _common_helpers import has_display, set_cycles_res

    use_cycles = not has_display()
except ImportError:
    use_cycles = False

extras = dict(engine="cycles") if use_cycles else {}


def test_lights():
    """ """
    bpy.ops.batoms.delete()
    r = Render(label="h2o")
    assert isinstance(r.lights, Lights)
    # properties
    r.lights.add("1")
    r.lights["1"].type = "POINT"
    assert r.lights["1"].type == "POINT"
    r.lights.add("2")
    r.lights.remove("2")
    assert len(r.lights) == 2


def test_light_direction():
    bpy.ops.batoms.delete()
    h2o = molecule("h2o", "H2O")
    h2o.render.lights["Default"].type = "POINT"
    h2o.render.lights["Default"].energy = 1000
    assert h2o.render.lights["Default"].type == "POINT"
    if use_cycles:
        set_cycles_res(h2o)
    h2o.get_image([0, 0, 1], output="light-direction.png", **extras)



import bpy


def test_create_node_group():
    from batoms.utils.butils import get_node_with_node_tree_by_name

    parent = bpy.data.node_groups.new("node_group1", "GeometryNodeTree")
    node = get_node_with_node_tree_by_name(parent.nodes, "node_group2")
    node.node_tree.nodes.new("GeometryNodeInputNamedAttribute")
    assert len(parent.nodes) == 1
    # default input and output nodes
    assert len(node.node_tree.nodes) == 3



from . import (
    ops_io,
    add_nanoparticle,
    add_nanotube,
    add_nanoribbon,
    add_object,
    add_surface,
    molecule_edit_atom,
    molecule_edit_bond,
    ops_batoms,
    ops_species,
    ops_volumetric_data,
    measure,
    manual_mapping,
    ops_render,
)

classes = [
    add_object.AddMolecule,
    add_object.AddBulk,
    add_object.AddAtoms,
    add_object.AddSmiles,
    add_surface.BuildSurfaceFCC100,
    add_surface.BuildSurfaceFCC110,
    add_surface.BuildSurfaceFCC111,
    add_surface.BuildSurfaceFCC211,
    add_surface.BuildSurfaceFCC111Root,
    add_surface.BuildSurfaceBCC100,
    add_surface.BuildSurfaceBCC110,
    add_surface.BuildSurfaceBCC111,
    add_surface.BuildSurfaceBCC111Root,
    add_surface.BuildSurfaceHCP0001,
    add_surface.BuildSurfaceHCP10m10,
    add_surface.BuildSurfaceHCP0001Root,
    add_surface.BuildSurfaceDiamond100,
    add_surface.BuildSurfaceDiamond111,
    add_nanotube.BuildNanotube,
    add_nanoribbon.BuildNanoribbon,
    add_nanoparticle.BuildDecahedron,
    add_nanoparticle.BuildIcosahedron,
    add_nanoparticle.BuildOctahedron,
    molecule_edit_atom.MolecueEditElement,
    molecule_edit_bond.MolecueEditBond,
    ops_batoms.BatomsReplace,
    ops_batoms.BatomModify,
    ops_batoms.ApplyCell,
    ops_batoms.ApplyTransform,
    ops_batoms.ApplyBoundary,
    ops_batoms.ApplyModelStyle,
    ops_batoms.ApplyRadiusStyle,
    ops_batoms.ApplyColorStyle,
    ops_batoms.BatomsCopy,
    ops_batoms.ApplyModelStyleSelected,
    ops_batoms.ApplyLabel,
    ops_batoms.BatomsAutoBuildSpecies,
    ops_batoms.AddSurface,
    ops_batoms.AddRootSurface,
    ops_batoms.BatomsJoin,
    ops_batoms.BatomsSeparate,
    ops_batoms.deleteBatoms,
    ops_batoms.deleteSelectedBatoms,
    ops_batoms.BatomsDeleteSelectedAtoms,
    ops_io.IMPORT_OT_batoms,
    ops_io.EXPORT_OT_batoms,
    ops_species.SpeciesAdd,
    ops_species.SpeciesRemove,
    ops_species.SpeciesUpdate,
    ops_species.SpeciesModify,
    ops_volumetric_data.VolumetricDataAdd,
    ops_volumetric_data.VolumetricDataRemove,
    ops_volumetric_data.VolumetricDataCreate,
    ops_render.RenderAdd,
    measure.MeasureButton,
]

manuals = [manual_mapping.batoms_ase_manual_map, manual_mapping.batoms_manual_map]


def register_class():
    from bpy.utils import register_class

    for cls in classes:
        register_class(cls)


def unregister_class():
    from bpy.utils import unregister_class

    for cls in reversed(classes):
        unregister_class(cls)


def register_manual_map():
    from bpy.utils import register_manual_map

    for cls in manuals:
        register_manual_map(cls)


def unregister_manual_map():
    from bpy.utils import unregister_manual_map

    for cls in reversed(manuals):
        unregister_manual_map(cls)


def register_menu():
    import bpy

    bpy.types.TOPBAR_MT_file_import.append(ops_io.menu_func_import_batoms)
    bpy.types.TOPBAR_MT_file_export.append(ops_io.menu_func_export_batoms)


def unregister_menu():
    import bpy

    bpy.types.TOPBAR_MT_file_import.remove(ops_io.menu_func_import_batoms)
    bpy.types.TOPBAR_MT_file_export.remove(ops_io.menu_func_export_batoms)



from bpy.props import BoolProperty, StringProperty
from batoms import Batoms
from batoms.ops.base import OperatorBatoms


class CavityAdd(OperatorBatoms):
    bl_idname = "surface.cavity_add"
    bl_label = "Add Cavity"
    bl_description = "Add cavity to a Batoms"

    name: StringProperty(
        name="name", default="0", description="Name of cavity to be added"
    )

    def execute(self, context):
        obj = context.object
        batoms = Batoms(label=context.object.batoms.label)
        batoms.cavity.settings.add(self.name)
        context.view_layer.objects.active = obj
        return {"FINISHED"}


class CavityRemove(OperatorBatoms):
    bl_idname = "surface.cavity_remove"
    bl_label = "Remove Cavity"
    bl_description = "Remove cavity to a Batoms"

    name: StringProperty(
        name="name", default="1-1-1", description="Name of cavity to be removed"
    )

    all: BoolProperty(name="all", default=False, description="Remove all cavitys")

    def execute(self, context):
        obj = context.object
        batoms = Batoms(label=obj.batoms.label)
        batoms.cavity.settings.remove((self.name))
        context.view_layer.objects.active = obj
        return {"FINISHED"}


class CavityDraw(OperatorBatoms):
    bl_idname = "surface.cavity_draw"
    bl_label = "Draw Cavity"
    bl_description = "Draw cavity to a Batoms"

    name: StringProperty(
        name="name", default="ALL", description="Name of cavity to be drawed"
    )

    def execute(self, context):
        obj = context.object
        batoms = Batoms(label=obj.batoms.label)
        batoms.cavity.draw()
        context.view_layer.objects.active = batoms.obj
        return {"FINISHED"}



"""
"""
from batoms.base.collection import Setting
import logging

logger = logging.getLogger(__name__)


class IsosurfaceSettings(Setting):
    def __init__(self, label, batoms=None, parent=None, isosurfacesetting=None) -> None:
        """IsosurfaceSetting object

            The IsosurfaceSetting object store the isosurface information.

            Parameters:

            label: str
                The label define the batoms object that a Setting belong to.


        Args:
            label (_type_):
                _description_
            batoms (_type_, optional):
                _description_. Defaults to None.
            isosurfacesetting (_type_, optional):
                _description_. Defaults to None.
        """
        Setting.__init__(self, label, coll_name="%s" % label)
        self.label = label
        self.batoms = batoms
        self.parent = parent
        self.name = "Bisosurface"
        # add a default level
        if isosurfacesetting is not None:
            for key, data in isosurfacesetting.items():
                self[key] = data

    def add(self, isosurface):
        if isinstance(isosurface, str):
            self[isosurface] = {}
        elif isinstance(isosurface, dict):
            self[isosurface["name"]] = isosurface

    def remove_isosurfaces(self, isosurface):
        for key in isosurface:
            name = "%s-%s" % (key[0], key[1])
            i = self.bpy_setting.find(name)
            if i != -1:
                self.bpy_setting.remove(i)

    def __repr__(self) -> str:
        s = "-" * 60 + "\n"
        s = "Center     level                color  \n"
        for iso in self.bpy_setting:
            s += "{0:10s}   {1:1.6f}  ".format(iso.name, iso.level)
            s += "[{:1.2f}  {:1.2f}  {:1.2f}  {:1.2f}] \n".format(
                iso.color[0], iso.color[1], iso.color[2], iso.color[3]
            )
        s += "-" * 60 + "\n"
        return s



import bpy
import pytest
import os

try:
    from _common_helpers import has_display, set_cycles_res

    use_cycles = not has_display()
except ImportError:
    use_cycles = False

extras = dict(engine="cycles") if use_cycles else {}
skip_test = bool(os.environ.get("NOTEST_CUBE", 0))


def test_crystal_shape(au):
    au.crystal_shape.settings[(1, 1, 1)] = {
        "distance": 3,
        "crystal": True,
        "symmetry": True,
    }
    au.crystal_shape.settings[(0, 0, 1)] = {
        "distance": 3,
        "crystal": True,
        "symmetry": True,
    }
    au.crystal_shape.draw(origin=au.cell.center)
    if use_cycles:
        set_cycles_res(au)
    au.get_image([0, 0, 1], output="plane-crystal.png", **extras)


def test_crystal_shape_ops(au):
    # need spglib to get the symmetry
    pytest.importorskip("spglib")
    bpy.context.view_layer.objects.active = au.obj
    bpy.ops.plane.crystal_shape_add(indices=(1, 1, 1))
    au.crystal_shape.settings["1-1-1"].symmetry = True
    au.crystal_shape.settings["1-1-1"].distance = 3.0
    bpy.ops.plane.crystal_shape_add(indices=(1, 0, 0))
    assert len(au.crystal_shape.settings) == 2
    bpy.ops.plane.crystal_shape_draw()
    bpy.ops.plane.crystal_shape_remove(name="1-0-0")
    assert len(au.crystal_shape.settings) == 8
    bpy.ops.plane.crystal_shape_draw()


def test_settings(au):
    """key search"""
    bpy.context.view_layer.objects.active = au.obj
    au.crystal_shape.settings.add((1, 1, 1))
    assert au.crystal_shape.settings.find((1, 1, 1)) is not None
    assert au.crystal_shape.settings.find("1-1-1") is not None
    au.crystal_shape.settings.remove((1, 1, 1))
    assert au.crystal_shape.settings.find((1, 1, 1)) is None


def test_gui(au):
    """crystal shape panel"""
    bpy.context.view_layer.objects.active = au.obj
    assert bpy.context.scene.Bcrystalshape.show is True
    bpy.context.scene.Bcrystalshape.show = False
    assert au.crystal_shape.show is False


def test_crystalshape_uilist(au):
    """crystalshape panel"""
    bpy.context.view_layer.objects.active = au.obj
    assert au.crystal_shape.settings.ui_list_index == 0
    bpy.ops.plane.crystal_shape_add(indices=(1, 1, 1))
    bpy.ops.plane.crystal_shape_add(indices=(1, 0, 0))
    assert au.coll.Bcrystalshape.ui_list_index == 1



import bpy
from batoms import Batoms


def test_bond_settings(ch4):
    bpy.context.view_layer.objects.active = ch4.obj
    bpy.ops.bond.bond_pair_remove(name="C-H")
    assert len(ch4.bond.settings) == 1
    bpy.ops.bond.bond_pair_add(species1="C", species2="H")
    assert len(ch4.bond.settings) == 2
    bpy.ops.bond.draw()


def test_bond_hydrogen_bond():
    bpy.ops.batoms.delete()
    bpy.ops.batoms.molecule_add(label="ch3oh", formula="CH3OH")
    ch3oh = Batoms("ch3oh")
    assert len(ch3oh.bond.bondlists) == 5
    bpy.ops.bond.show_hydrogen_bond()
    assert len(ch3oh.bond.bondlists) == 8


def test_bond_search_bond(tio2):
    tio2.boundary = 0.01
    tio2.model_style = 1
    assert tio2.bond.show_search is False
    bpy.ops.bond.show_search()
    assert tio2.bond.show_search is True



from batoms.render.render import Render
from batoms.render.light import Light, Lights
from batoms.render.camera import Camera


from . import (
    bpy_data,
)

__all__ = ["Render", "Light", "Lights", "Camera"]

classes = [
    # internal data first
    bpy_data.LightSetting,
    bpy_data.CameraSetting,
    bpy_data.Render,
    # ops.ProteinAdd,
    # ops.ProteinRemove,
    # ops.ProteinDraw,
    # ui_list.BATOMS_MT_sheet_context_menu,
    # ui_list.BATOMS_UL_sheet,
    # ui_list.BATOMS_PT_sheet,
    # ui_list.BATOMS_MT_helix_context_menu,
    # ui_list.BATOMS_UL_helix,
    # ui_list.BATOMS_PT_helix,
    # ui_list.BATOMS_MT_turn_context_menu,
    # ui_list.BATOMS_UL_turn,
    # ui_list.BATOMS_PT_turn,
]


def register_class():
    from bpy.types import Collection, Object
    from bpy.props import PointerProperty
    from bpy.utils import register_class

    for cls in classes:
        register_class(cls)
    # attach to blender internal data
    Collection.Brender = PointerProperty(name="Brender", type=bpy_data.Render)
    Object.Blight = PointerProperty(name="Blight", type=bpy_data.LightSetting)
    Object.Bcamera = PointerProperty(name="Bcamera", type=bpy_data.CameraSetting)


def unregister_class():
    from bpy.types import Collection, Object
    from bpy.utils import unregister_class

    for cls in reversed(classes):
        unregister_class(cls)

    del Collection.Brender
    del Object.Blight
    del Object.Bcamera



import bmesh
from bpy.types import Operator
from bpy.props import BoolProperty, FloatProperty, StringProperty
from batoms import Batoms
from batoms.ops.base import OperatorBatoms


class SpeciesAdd(OperatorBatoms):
    bl_idname = "batoms.species_add"
    bl_label = "Add Species"
    bl_description = "Add Species to a Batoms"

    species: StringProperty(
        name="species", default="", description="Species to be added"
    )

    def execute(self, context):
        if self.species == "":
            return {"FINISHED"}
        obj = context.object
        batoms = Batoms(label=context.object.batoms.label)
        batoms.species.add(self.species)
        context.view_layer.objects.active = obj
        self.report({"INFO"}, "{} is added.".format(self.species))
        return {"FINISHED"}


class SpeciesRemove(OperatorBatoms):
    bl_idname = "batoms.species_remove"
    bl_label = "Remove Species"
    bl_description = "Remove Species to a Batoms"

    species: StringProperty(
        name="species", default="C", description="Species to be removed"
    )

    all: BoolProperty(name="all", default=False, description="Remove all Speciess")

    def execute(self, context):
        obj = context.object
        batoms = Batoms(label=obj.batoms.label)
        batoms.species.remove((self.species))
        context.view_layer.objects.active = obj
        self.report({"INFO"}, "{} is removed.".format(self.species))
        return {"FINISHED"}


class SpeciesUpdate(OperatorBatoms):
    bl_idname = "batoms.species_update"
    bl_label = "Update Species"
    bl_description = "Update Species to a Batoms"

    def execute(self, context):
        obj = context.object
        batoms = Batoms(label=obj.batoms.label)
        batoms.species.update()
        context.view_layer.objects.active = batoms.obj
        return {"FINISHED"}


class SpeciesModify(Operator):
    bl_idname = "batoms.species_modify"
    bl_label = "Modify Species"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Modify Species"

    key: StringProperty(
        name="key", default="style", description="Replaced by this species"
    )

    slice: BoolProperty(
        name="slice",
        default=False,
    )
    boundary: BoolProperty(
        name="boundary",
        default=False,
    )
    distance: FloatProperty(
        name="distance", description="Distance from origin", default=1
    )

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj:
            return obj.batoms.type == "MS" and obj.mode == "EDIT"
        else:
            return False

    def execute(self, context):
        obj = context.object
        data = obj.data
        bm = bmesh.from_edit_mesh(data)
        v = [s.index for s in bm.select_history if isinstance(s, bmesh.types.BMVert)]
        batoms = Batoms(label=obj.batoms.label)
        for i in v:
            setattr(batoms.bonds[i], self.key, getattr(self, self.key))
        # batoms.draw()
        return {"FINISHED"}



import bpy
from batoms.ribbon.profile import ellipse, rectangle
from batoms.base.collection import Setting
from time import time
import numpy as np
import logging

logger = logging.getLogger(__name__)


class Protein:
    def __init__(
        self,
        batoms=None,
        chains={},
        residues=[],
        sheets={},
        helixs={},
        turns={},
    ) -> None:
        self.batoms = batoms
        self.sheetsetting = SheetSetting(self.batoms.label, batoms)
        self.helixsetting = HelixSetting(self.batoms.label, batoms)
        self.chains = chains
        self.residues = residues
        self.sheets = sheets
        self.helixs = helixs
        self.turns = turns

    def setting(self, datas):
        if datas is None:
            return
        if "sheet" in datas:
            self.sheetsetting.from_dict(datas["sheet"])
        if "helix" in datas:
            self.helixsetting.from_dict(datas["helix"])

    def update(self):
        tstart = time()
        arrays = self.batoms.arrays
        if "types" not in arrays:
            return
        # build Chains
        chains = {}
        chainIDs = np.unique(arrays["chainids"])
        for chainID in chainIDs:
            chain = Chain(chainID)
            indices = np.where(arrays["chainids"] == chainID)[0]
            chain.indices = indices
            chains[chainID] = chain
        # build residues
        residues = []
        # get Sheet
        names = np.core.defchararray.add(
            arrays["chainids"], arrays["residuenumbers"].astype("U20")
        )
        residuenames, indices = np.unique(names, return_index=True)
        residuenames = residuenames[np.argsort(indices)]
        indices = indices[np.argsort(indices)]
        lastResi = np.where(names == residuenames[-1])[0][-1]
        indices = np.append(indices, lastResi + 1)
        # find real residue and its Ca, C, O
        nresi = len(residuenames)
        for i in range(nresi):
            resi = Residue(
                names[indices[i]],
                arrays["residuenames"][indices[i]],
                arrays["chainids"][indices[i]],
                arrays["residuenumbers"][indices[i]],
            )
            resi.indices = list(range(indices[i], indices[i + 1]))
            if "ATOM" in arrays["types"]:
                for j in range(indices[i], indices[i + 1]):
                    if arrays["atomtypes"][j] == "CA":
                        resi.Ca = j
                    elif arrays["atomtypes"][j] == "C":
                        resi.C = j
                    elif arrays["atomtypes"][j] == "O":
                        resi.O = j  # noqa E741
            else:
                resi.type = -1
            residues.append(resi)
        # calculate the plane
        for i in range(nresi - 1):
            if residues[i].type == -1 or residues[i + 1].type == -1:
                continue
            plane = GetPeptidePlane(residues[i], residues[i + 1], arrays["positions"])
            residues[i].plane = plane
        # init sheets
        sheets = {}
        for sheet in self.sheetsetting.bpy_setting:
            sheet1 = Sheet(
                sheet.name,
                1,
                sheet.startChain,
                sheet.endChain,
                sheet.startResi,
                sheet.endResi,
            )
            sheet1.color = sheet.color
            sheets[sheet.name] = sheet1
        # init helixs
        helixs = {}
        for helix in self.helixsetting.bpy_setting:
            helix1 = Helix(
                helix.name,
                2,
                helix.startChain,
                helix.endChain,
                helix.startResi,
                helix.endResi,
            )
            helix1.color = helix.color
            helixs[helix.name] = helix1
        for i in range(nresi - 1):
            if residues[i].type == -1 or residues[i + 1].type == -1:
                continue
            # add sheet
            for sheet in self.sheetsetting:
                if sheet.startChain == residues[i].chainID:
                    if residues[i].resSeq == sheet.startResi and i > 0:
                        residues[i].type = 1
                        sheets[sheet.name].append(residues[i - 1])
                        sheets[sheet.name].append(residues[i])
                        break
                    elif (
                        residues[i].resSeq > sheet.startResi
                        and residues[i].resSeq < sheet.endResi  # noqa W503
                    ):
                        residues[i].type = 1
                        sheets[sheet.name].append(residues[i])
                        break
                    elif residues[i].resSeq == sheet.endResi and i < nresi - 1:
                        residues[i].type = 1
                        sheets[sheet.name].append(residues[i])
                        sheets[sheet.name].append(residues[i + 1])
                        break
            # add helix
            for helix in self.helixsetting:
                if helix.startChain == residues[i].chainID:
                    if residues[i].resSeq == helix.startResi and i > 0:
                        residues[i].type = 2
                        helixs[helix.name].append(residues[i - 1])
                        helixs[helix.name].append(residues[i])
                        break
                    elif (
                        residues[i].resSeq > helix.startResi
                        and residues[i].resSeq < helix.endResi  # noqa W503
                    ):
                        residues[i].type = 2
                        helixs[helix.name].append(residues[i])
                        break
                    elif residues[i].resSeq == helix.endResi and i < nresi - 1:
                        residues[i].type = 2
                        helixs[helix.name].append(residues[i])
                        helixs[helix.name].append(residues[i + 1])
                        break
        # init turns
        turns = {}
        if residues[0].type == 0:
            turn = Turn(
                residues[0].name,
                0,
                residues[0].chainID,
                residues[0].resSeq,
            )
            turns[turn.name] = turn
            turns[turn.name].append(residues[0])
        # add turn
        for i in range(1, nresi - 1):
            if residues[i - 1].chainID != turn.startChain or (
                residues[i - 1].type != 0 and residues[i].type == 0
            ):
                turn = Turn(
                    residues[i].name,
                    0,
                    residues[i].chainID,
                    residues[i].resSeq,
                )
                turns[turn.name] = turn
                if i > 1:
                    turns[turn.name].append(residues[i - 2])
                turns[turn.name].append(residues[i - 1])
                turns[turn.name].append(residues[i])
            elif residues[i].type == 0 and residues[i + 1].type == 0:
                turns[turn.name].append(residues[i])
            elif residues[i].type == 0 and residues[i + 1].type != 0:
                turns[turn.name].append(residues[i])
                if i < nresi - 1:
                    turns[turn.name].append(residues[i + 1])
                    if i < nresi - 2:
                        turns[turn.name].append(residues[i + 2])

        self.residues = residues
        self.chains = chains
        self.sheets = sheets
        self.helixs = helixs
        self.turns = turns
        self.batoms.selects["all"].show = False
        # HETATM
        indices = np.where(arrays["types"] == "HETATM")[0]
        self.batoms.selects.add("heta", indices)
        logger.debug("update ribbon: %s" % (time() - tstart))


class Chain:
    def __init__(self, chainID, residues=[]) -> None:
        self.chainID = chainID
        self.residues = residues
        self.indices = []


class Residue:
    """
    Type:
        0:
        1: Sheet
        2: Helix
        -1: water or ...
    """

    def __init__(self, name, resName, chainID, resSeq, type=0) -> None:
        self.name = name
        self.resName = resName
        self.chainID = chainID
        self.resSeq = resSeq
        self.Ca = 0
        self.C = 0
        self.O = 0  # noqa E741
        self.plane = None
        self.type = type
        self.indices = []


class secondaryStructure:
    """
    Type:
        0:
        1: Sheet
        2: Helix
        -1: water or ...
    """

    def __init__(
        self,
        name,
        type,
        startChain,
        endChain,
        startResi,
        endResi,
    ) -> None:
        self.name = name
        self.type = type
        self.startChain = startChain
        self.endChain = endChain
        self.startResi = startResi
        self.endResi = endResi
        self.residues = []
        self.indices = []
        self.Ca = []
        self.C = []
        self.O = []  # noqa E741
        self.extrude = 1.0
        self.depth = 0.25
        self.resolution = 20

    def append(self, residue):
        self.residues.append(residue)
        self.indices.extend(residue.indices)
        self.Ca.append(residue.Ca)
        self.C.append(residue.C)
        self.O.append(residue.O)

    @property
    def positions(self):
        n = len(self.residues)
        positions = np.ones((n, 3))
        for i in range(n):
            positions[i] = self.residues[i].plane["position"]
        return positions

    @property
    def profiles(self):
        profiles = rectangle(self.extrude, self.depth)
        return profiles

    @property
    def sides(self):
        n = len(self.residues)
        sides = np.ones((n, 3))
        for i in range(n):
            sides[i] = self.residues[i].plane["side"]
        return sides

    @property
    def normals(self):
        n = len(self.residues)
        normals = np.ones((n, 3))
        for i in range(n):
            normals[i] = self.residues[i].plane["normal"]
        return normals

    @property
    def tilts(self):
        n = len(self.residues)
        tilts = np.zeros(n)
        for i in range(n):
            tilts[i] = np.arccos(self.residues[i].plane["side"][2])
        return tilts

    @property
    def scales(self):
        n = len(self.residues)
        scales = np.ones((self.resolution * (n - 1), 3))
        scales[-self.resolution :, 0] *= np.linspace(1.5, 0.0, self.resolution)
        return scales

    def as_dict(self):
        """
         A complication arises when the direction of the carbonyl
        oxygen flips, as is always the case between adjacent
        residues of B sheets.
        """
        n = len(self.residues)
        for i in range(1, n):
            if (
                np.dot(
                    self.residues[i - 1].plane["side"], self.residues[i].plane["side"]
                )
                < 0  # noqa W503
            ):
                self.residues[i].plane["flipped"] = True
                self.residues[i].plane["normal"] = -self.residues[i].plane["normal"]
                self.residues[i].plane["side"] = -self.residues[i].plane["side"]
        data = {
            "vertices": self.positions,
            "color": self.color,
            "tilts": self.tilts,
            "sides": self.sides,
            "normals": self.normals,
            "profiles": self.profiles,
            "scales": self.scales,
            "depth": self.depth,
            "resolution": self.resolution,
        }
        return data


class Sheet(secondaryStructure):
    def __init__(
        self,
        name,
        type,
        startChain,
        endChain,
        startResi,
        endResi,
    ) -> None:
        secondaryStructure.__init__(
            self,
            name,
            type,
            startChain,
            endChain,
            startResi,
            endResi,
        )
        self.type = 1


class Helix(secondaryStructure):
    """ """

    def __init__(
        self,
        name,
        type,
        startChain,
        endChain,
        startResi,
        endResi,
    ) -> None:
        secondaryStructure.__init__(
            self,
            name,
            type,
            startChain,
            endChain,
            startResi,
            endResi,
        )
        self.type = 2

    @property
    def positions(self):
        n = len(self.residues)
        positions = np.ones((n, 3))
        for i in range(n):
            positions[i] = (
                self.residues[i].plane["position"]
                + self.residues[i].plane["normal"] * 1.5  # noqa W503
            )
        return positions

    @property
    def profiles(self):
        profiles = ellipse(32, self.extrude, self.depth)
        return profiles

    @property
    def scales(self):
        n = len(self.residues)
        scales = np.ones((self.resolution * (n - 1), 3))
        return scales


class Turn(secondaryStructure):
    """ """

    def __init__(
        self,
        name,
        type,
        startChain=None,
        endChain=None,
        startResi=None,
        endResi=None,
    ) -> None:
        secondaryStructure.__init__(
            self,
            name,
            type,
            startChain,
            endChain,
            startResi,
            endResi,
        )
        self.type = 2
        self.extrude = 1.0
        self.depth = 0.3
        self.color = [0, 1, 0, 1]

    @property
    def positions(self):
        n = len(self.residues)
        positions = np.ones((n, 3))
        for i in range(n):
            if self.residues[i].type == 2:
                positions[i] = (
                    self.residues[i].plane["position"]
                    + self.residues[i].plane["normal"] * 1.5  # noqa W503
                )
            else:
                positions[i] = self.residues[i].plane["position"]
        return positions

    def as_dict(self):
        data = {
            "vertices": self.positions,
            "color": self.color,
            "tilts": self.tilts,
            "extrudes": self.extrude,
            "radius": self.depth,
        }
        return data


class SheetSetting(Setting):
    """
    sheet = {
    'name': '%s-%s'%(chainId, sheetId),
    'startChain': startChain,
    'startResi': startResi,
    'endChain': endChain,
    'endResi': endResi,
    }
    """

    def __init__(
        self,
        label,
        batoms=None,
    ) -> None:
        Setting.__init__(self, label, coll_name=label)
        self.label = label
        self.name = "Bprotein"
        self.batoms = batoms
        self.sheet_datas = {}

    def get_bpy_setting(self):
        if self.coll_name:
            coll = bpy.data.collections.get(self.coll_name)
            data = getattr(coll, self.name)
        else:
            raise KeyError("The collection property {} not exist!".format(self.name))
        return data.settings_sheet

    @property
    def show(self):
        return self.get_show()

    @show.setter
    def show(self, state):
        self.set_show(state)

    def get_show(self):
        return self.batoms.attributes["show"][self.indices]

    def set_show(self, show, only_atoms=True):
        show0 = self.batoms.show
        show0[self.indices] = show
        self.batoms.set_attributes({"show": show0})


class HelixSetting(Setting):
    """ """

    def __init__(
        self,
        label,
        batoms=None,
    ) -> None:
        Setting.__init__(self, label, coll_name=label)
        self.label = label
        self.name = "Bprotein"
        self.batoms = batoms
        self.helix_datas = {}

    def get_bpy_setting(self):
        if self.coll_name:
            coll = bpy.data.collections.get(self.coll_name)
            data = getattr(coll, self.name)
        else:
            raise KeyError("The collection property {} not exist!".format(self.name))
        return data.settings_helix


def GetPeptidePlane(resi1, resi2, positions):
    """
    Peptide plane: The atoms of the group, O=C-N-H,
    are fixed on the same plane, known as the peptide plane.
    The whole plane may rotate around the N-Cα bond (φ angle)
    or C-Cα bond (ψ angle).

    Cα is the carbon atom connected to the R group.

    In our case, we use the -Cα-O-Cα- plane as PetidePlane
    """
    # get GetPeptidePlane
    v1 = positions[resi2.Ca] - positions[resi1.Ca]
    v2 = positions[resi1.O] - positions[resi1.Ca]
    forward = v1 / np.linalg.norm(v1)
    position = (positions[resi1.Ca] + positions[resi2.Ca]) / 2
    # position = positions[resi1.Ca]
    normal = np.cross(v1, v2)
    normal = normal / np.linalg.norm(normal)
    side = np.cross(normal, forward)
    # tilt = np.arccos(side[2])
    flipped = False
    plane = {
        "position": position,
        "forward": forward,
        "normal": normal,
        "side": side,
        "flipped": flipped,
    }
    return plane


def GetBackbone():
    """
    the repeating -Cα-C-N-Cα- is the backbone of the peptide chain.
    """
